# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import itertools

from EIVideo.paddlevideo.utils import get_logger

logger = get_logger("paddlevideo")
"""
Implement precise bn, which is useful for improving accuracy.
"""


@paddle.no_grad()  # speed up and save CUDA memory
def do_preciseBN(model, data_loader, parallel, num_iters=200):
    """
    Recompute and update the batch norm stats to make them more precise. During
    training both BN stats and the weight are changing after every iteration, so
    the running average can not precisely reflect the actual stats of the
    current model.
    In this function, the BN stats are recomputed with fixed weights, to make
    the running average more precise. Specifically, it computes the true average
    of per-batch mean/variance instead of the running average.
    This is useful to improve validation accuracy.
    Args:
        model: the model whose bn stats will be recomputed
        data_loader: an iterator. Produce data as input to the model
        num_iters: number of iterations to compute the stats.
    Return:
        the model with precise mean and variance in bn layers.
    """
    bn_layers_list = [
        m for m in model.sublayers()
        if any((isinstance(m, bn_type)
                for bn_type in (paddle.nn.BatchNorm1D, paddle.nn.BatchNorm2D,
                                paddle.nn.BatchNorm3D))) and m.training
    ]
    if len(bn_layers_list) == 0:
        return

    # moving_mean=moving_mean*momentum+batch_mean*(1.−momentum)
    # we set momentum=0. to get the true mean and variance during forward
    momentum_actual = [bn._momentum for bn in bn_layers_list]
    for bn in bn_layers_list:
        bn._momentum = 0.

    running_mean = [paddle.zeros_like(bn._mean)
                    for bn in bn_layers_list]  #pre-ignore
    running_var = [paddle.zeros_like(bn._variance) for bn in bn_layers_list]

    ind = -1
    for ind, data in enumerate(itertools.islice(data_loader, num_iters)):
        logger.info("doing precise BN {} / {}...".format(ind + 1, num_iters))
        if parallel:
            model._layers.train_step(data)
        else:
            model.train_step(data)

        for i, bn in enumerate(bn_layers_list):
            # Accumulates the bn stats.
            running_mean[i] += (bn._mean - running_mean[i]) / (ind + 1)
            running_var[i] += (bn._variance - running_var[i]) / (ind + 1)

    assert ind == num_iters - 1, (
        "update_bn_stats is meant to run for {} iterations, but the batch_sampler stops at {} iterations."
        .format(num_iters, ind))

    # Sets the precise bn stats.
    for i, bn in enumerate(bn_layers_list):
        bn._mean.set_value(running_mean[i])
        bn._variance.set_value(running_var[i])
        bn._momentum = momentum_actual[i]



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import os.path as osp
import time

from tqdm import tqdm
import paddle
import paddle.nn.functional as F
from EIVideo.paddlevideo.utils import get_logger
from EIVideo.paddlevideo.utils import main_only


def pretrain_vit_param_trans(model, state_dicts, num_patches, seg_num,
                             attention_type):
    """
    Convert ViT's pre-trained model parameters to a parameter dictionary that matches the existing model
    """
    if 'head' + '.weight' in state_dicts:
        del state_dicts['head' + '.weight']
    if 'head' + '.bias' in state_dicts:
        del state_dicts['head' + '.bias']

    total_len = len(model.state_dict())
    if num_patches + 1 != state_dicts['pos_embed'].shape[1]:
        pos_embed = state_dicts['pos_embed']
        cls_pos_embed = pos_embed[0, 0, :].unsqueeze(0).unsqueeze(1)
        other_pos_embed = pos_embed[0,
                                    1:, :].unsqueeze(0).unsqueeze(1).transpose(
                                        (0, 1, 3, 2))
        new_pos_embed = F.interpolate(other_pos_embed,
                                      size=(other_pos_embed.shape[-2],
                                            num_patches),
                                      mode='nearest')
        new_pos_embed = new_pos_embed.squeeze(0).transpose((0, 2, 1))
        new_pos_embed = paddle.concat((cls_pos_embed, new_pos_embed), axis=1)
        state_dicts['pos_embed'] = new_pos_embed
        time.sleep(0.01)

    if 'time_embed' in state_dicts and seg_num != state_dicts[
            'time_embed'].shape[1]:
        time_embed = state_dicts['time_embed'].transpose(
            (0, 2, 1)).unsqueeze(0)
        new_time_embed = F.interpolate(time_embed,
                                       size=(time_embed.shape[-2], seg_num),
                                       mode='nearest')
        state_dicts['time_embed'] = new_time_embed.squeeze(0).transpose(
            (0, 2, 1))
        time.sleep(0.01)
    with tqdm(total=total_len,
              position=1,
              bar_format='{desc}',
              desc="Loading weights") as desc:
        if attention_type == 'divided_space_time':
            new_state_dicts = state_dicts.copy()
            for key in tqdm(state_dicts):
                if 'blocks' in key and 'attn' in key:
                    desc.set_description("Loading %s" % key)
                    new_key = key.replace('attn', 'temporal_attn')
                    if not new_key in state_dicts:
                        new_state_dicts[new_key] = state_dicts[key]
                    else:
                        new_state_dicts[new_key] = state_dicts[new_key]
                if 'blocks' in key and 'norm1' in key:
                    desc.set_description("Loading %s" % key)
                    new_key = key.replace('norm1', 'temporal_norm1')
                    if not new_key in state_dicts:
                        new_state_dicts[new_key] = state_dicts[key]
                    else:
                        new_state_dicts[new_key] = state_dicts[new_key]
                time.sleep(0.01)
    ret_str = "loading {:<20d} weights completed.".format(
        len(model.state_dict()))
    desc.set_description(ret_str)
    return new_state_dicts


def pretrain_resnet18_param_trans(model, loaded_dict):
    encoder_dict = model.encoder.state_dict()
    pose_encoder_dict = model.pose_encoder.state_dict()

    names = ['encoder.', 'encoder_day.', 'encoder_night.']
    for name in names:
        for key, value in loaded_dict.items():
            key = str(name + key)
            if key in encoder_dict:
                encoder_dict[key] = value

    num_input_images = 2
    loaded_dict['conv1.weight'] = paddle.concat(
        [loaded_dict['conv1.weight']] * num_input_images, 1) / num_input_images

    for name, value in loaded_dict.items():
        name = str('encoder.' + name)
        if name in pose_encoder_dict:
            pose_encoder_dict[name] = value

    return encoder_dict, pose_encoder_dict


#XXX(shipping): maybe need load N times because of different cards have different params.
@main_only
def load_ckpt(model, weight_path, **kargs):
    """
    1. Load pre-trained model parameters
    2. Extract and convert from the pre-trained model to the parameters
    required by the existing model
    3. Load the converted parameters of the existing model
    """
    #model.set_state_dict(state_dict)

    if not osp.isfile(weight_path):
        raise IOError(f'{weight_path} is not a checkpoint file')
    #state_dicts = load(weight_path)

    logger = get_logger("paddlevideo")
    state_dicts = paddle.load(weight_path)
    if 'ResnetEncoder' in str(model):
        encoder_dict, pose_encoder_dict = pretrain_resnet18_param_trans(
            model, state_dicts)
        tmp = model.state_dict()
        tmp.update(
            {'backbone.encoder.' + k: v
             for (k, v) in encoder_dict.items()})
        tmp.update({
            'backbone.pose_encoder.' + k: v
            for (k, v) in pose_encoder_dict.items()
        })
    elif "VisionTransformer" in str(model):  # For TimeSformer case
        tmp = pretrain_vit_param_trans(model, state_dicts,
                                       kargs['num_patches'], kargs['seg_num'],
                                       kargs['attention_type'])
    else:
        tmp = {}
        total_len = len(model.state_dict())
        with tqdm(total=total_len,
                  position=1,
                  bar_format='{desc}',
                  desc="Loading weights") as desc:
            for item in tqdm(model.state_dict(), total=total_len, position=0):
                name = item
                desc.set_description('Loading %s' % name)
                if name not in state_dicts:  # Convert from non-parallel model
                    if str('backbone.' + name) in state_dicts:
                        tmp[name] = state_dicts['backbone.' + name]
                else:  # Convert from parallel model
                    tmp[name] = state_dicts[name]
                time.sleep(0.01)
        ret_str = "loading {:<20d} weights completed.".format(
            len(model.state_dict()))
        desc.set_description(ret_str)
    model.set_state_dict(tmp)


def mkdir(dir):
    if not os.path.exists(dir):
        # avoid error when train with multiple gpus
        try:
            os.makedirs(dir)
        except:
            pass


"""
def save(state_dicts, file_name):
    def convert(state_dict):
        model_dict = {}

        for k, v in state_dict.items():
            if isinstance(
                    v,
                (paddle.fluid.framework.Variable, paddle.fluid.core.VarBase)):
                model_dict[k] = v.numpy()
            else:
                model_dict[k] = v

        return model_dict

    final_dict = {}
    for k, v in state_dicts.items():
        if isinstance(
                v,
            (paddle.fluid.framework.Variable, paddle.fluid.core.VarBase)):
            final_dict = convert(state_dicts)
            break
        elif isinstance(v, dict):
            final_dict[k] = convert(v)
        else:
            final_dict[k] = v

    with open(file_name, 'wb') as f:
        pickle.dump(final_dict, f, protocol=2)
"""


@main_only
def save(obj, path):
    paddle.save(obj, path)


def load(file_name):
    if not osp.isfile(file_name):
        raise IOError(f'{file_name} not exist')
    return paddle.load(file_name)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import yaml
from EIVideo.paddlevideo.utils.logger import coloring, setup_logger

__all__ = ['get_config']

logger = setup_logger("./", name="paddlevideo", level="INFO")


class AttrDict(dict):
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key in self.__dict__:
            self.__dict__[key] = value
        else:
            self[key] = value


def create_attr_dict(yaml_config):
    from ast import literal_eval
    for key, value in yaml_config.items():
        if type(value) is dict:
            yaml_config[key] = value = AttrDict(value)
        if isinstance(value, str):
            try:
                value = literal_eval(value)
            except BaseException:
                pass
        if isinstance(value, AttrDict):
            create_attr_dict(yaml_config[key])
        else:
            yaml_config[key] = value


def parse_config(cfg_file):
    """Load a config file into AttrDict"""
    with open(cfg_file, 'r') as fopen:
        yaml_config = AttrDict(yaml.load(fopen, Loader=yaml.SafeLoader))
    create_attr_dict(yaml_config)
    return yaml_config


def print_dict(d, delimiter=0):
    """
    Recursively visualize a dict and
    indenting acrrording by the relationship of keys.
    """
    placeholder = "-" * 60
    for k, v in sorted(d.items()):
        if isinstance(v, dict):
            logger.info("{}{} : ".format(delimiter * " ", coloring(k,
                                                                   "HEADER")))
            print_dict(v, delimiter + 4)
        elif isinstance(v, list) and len(v) >= 1 and isinstance(v[0], dict):
            logger.info("{}{} : ".format(delimiter * " ",
                                         coloring(str(k), "HEADER")))
            for value in v:
                print_dict(value, delimiter + 4)
        else:
            logger.info("{}{} : {}".format(delimiter * " ",
                                           coloring(k, "HEADER"),
                                           coloring(v, "OKGREEN")))

        if k.isupper():
            logger.info(placeholder)


def print_config(config):
    """
    visualize configs
    Arguments:
        config: configs
    """
    print_dict(config)


def check_config(config):
    """
    Check config
    """
    pass


def override(dl, ks, v):
    """
    Recursively replace dict of list
    Args:
        dl(dict or list): dict or list to be replaced
        ks(list): list of keys
        v(str): value to be replaced
    """
    def str2num(v):
        try:
            return eval(v)
        except Exception:
            return v

    assert isinstance(dl, (list, dict)), ("{} should be a list or a dict")
    assert len(ks) > 0, ('lenght of keys should larger than 0')
    if isinstance(dl, list):
        k = str2num(ks[0])
        if len(ks) == 1:
            assert k < len(dl), ('index({}) out of range({})'.format(k, dl))
            dl[k] = str2num(v)
        else:
            override(dl[k], ks[1:], v)
    else:
        if len(ks) == 1:
            #assert ks[0] in dl, ('{} is not exist in {}'.format(ks[0], dl))
            if not ks[0] in dl:
                logger.warning('A new filed ({}) detected!'.format(ks[0], dl))
            dl[ks[0]] = str2num(v)
        else:
            assert ks[0] in dl, (
                '({}) doesn\'t exist in {}, a new dict field is invalid'.format(
                    ks[0], dl))
            override(dl[ks[0]], ks[1:], v)


def override_config(config, options=None):
    """
    Recursively override the config
    Args:
        config(dict): dict to be replaced
        options(list): list of pairs(key0.key1.idx.key2=value)
            such as: [
                epochs=20',
                'PIPELINE.train.transform.1.ResizeImage.resize_short=300'
            ]
    Returns:
        config(dict): replaced config
    """
    if options is not None:
        for opt in options:
            assert isinstance(opt,
                              str), ("option({}) should be a str".format(opt))
            assert "=" in opt, (
                "option({}) should contain a ="
                "to distinguish between key and value".format(opt))
            pair = opt.split('=')
            assert len(pair) == 2, ("there can be only a = in the option")
            key, value = pair
            keys = key.split('.')
            override(config, keys, value)

    return config


def get_config(fname, overrides=None, show=True):
    """
    Read config from file
    """
    assert os.path.exists(fname), ('config file({}) is not exist'.format(fname))
    config = parse_config(fname)
    override_config(config, overrides)
    if show:
        print_config(config)
    check_config(config)
    return config



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .registry import Registry
from .build_utils import build
from .config import *
from .logger import setup_logger, coloring, get_logger
from .record import AverageMeter, build_record, log_batch, log_epoch
from .dist_utils import get_dist_info, main_only
from .save_load import save, load, load_ckpt, mkdir
from .precise_bn import do_preciseBN
from .profiler import add_profiler_step
__all__ = ['Registry', 'build']



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import Registry

BACKBONES = Registry('backbone')
HEADS = Registry('head')
RECOGNIZERS = Registry('recognizer')
LOCALIZERS = Registry('localizer')
PARTITIONERS = Registry('partitioner')
SEGMENT = Registry('segmentation')
LOSSES = Registry('loss')
ROI_EXTRACTORS = Registry('roi_extractor')
DETECTORS = Registry('detectors')
BBOX_ASSIGNERS = Registry('bbox_assigner')
BBOX_SAMPLERS = Registry('bbox_sampler')
BBOX_CODERS = Registry('bbox_coder')
ESTIMATORS = Registry('estimator')
MULTIMODAL = Registry('multimodal')
SEGMENT = Registry('segment')



import paddle
import paddle.nn as nn

from ..registry import BACKBONES
from EIVideo.paddlevideo.modeling.backbones.aspp_manet import build_aspp
from EIVideo.paddlevideo.modeling.backbones.decoder_manet import build_decoder
from EIVideo.paddlevideo.modeling.backbones.resnet_manet import build_backbone


class FrozenBatchNorm2d(nn.Layer):
    def __init__(self, n):
        super(FrozenBatchNorm2d, self).__init__()
        self.register_buffer("weight", paddle.ones(n))
        self.register_buffer("bias", paddle.zeros(n))
        self.register_buffer("running_mean", paddle.zeros(n))
        self.register_buffer("running_var", paddle.ones(n))

    def forward(self, x):
        if x.dtype == paddle.float16:
            self.weight = self.weight.half()
            self.bias = self.bias.half()
            self.running_mean = self.running_mean.half()
            self.running_var = self.running_var.half()
        scale = self.weight * self.running_var.rsqrt()
        bias = self.bias - self.running_mean * scale
        scale = scale.reshape(1, -1, 1, 1)
        bias = bias.reshape(1, -1, 1, 1)
        return x * scale + bias


@BACKBONES.register()
class DeepLab(nn.Layer):
    def __init__(self,
                 backbone='resnet',
                 output_stride=16,
                 num_classes=21,
                 freeze_bn=False,
                 pretrained=None):
        super(DeepLab, self).__init__()
        if backbone == 'drn':
            output_stride = 8
        if freeze_bn == True:
            print("Use frozen BN in DeepLab")
            BatchNorm = FrozenBatchNorm2d
        else:
            BatchNorm = nn.BatchNorm2D

        self.backbone = build_backbone(output_stride, BatchNorm, pretrained)
        self.aspp = build_aspp(backbone, output_stride, BatchNorm)
        self.decoder = build_decoder(num_classes, backbone, BatchNorm)


    def forward(self, input):
        x, low_level_feat = self.backbone(input)
        x = self.aspp(x)
        x = self.decoder(x, low_level_feat)
        return x

    def freeze_bn(self):
        for m in self.sublayers():
            if isinstance(m, nn.BatchNorm2D):
                m.eval()

    def get_1x_lr_params(self):
        modules = [self.backbone]
        for i in range(len(modules)):
            for m in modules[i].named_modules():
                if isinstance(m[1], nn.Conv2D) or isinstance(
                        m[1], nn.BatchNorm2D):
                    for p in m[1].parameters():
                        if p.requires_grad:
                            yield p

    def get_10x_lr_params(self):
        modules = [self.aspp, self.decoder]
        for i in range(len(modules)):
            for m in modules[i].named_modules():
                if isinstance(m[1], nn.Conv2D) or isinstance(
                        m[1], nn.BatchNorm2D):
                    for p in m[1].parameters():
                        if p.requires_grad:
                            yield p


if __name__ == "__main__":
    model = DeepLab(backbone='resnet', output_stride=16)
    model.eval()
    input = paddle.rand([2, 3, 513, 513])
    output = model(input)
    print(output.shape)




# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

from abc import abstractmethod
from ... import builder
import paddle.nn as nn


class BaseSegment(nn.Layer):
    """Base class for semi-Video Object Segmentation.
    All subclass should overwrite:

    - Methods:``train_step``, supporting to forward when training.
    - Methods:``valid_step``, supporting to forward when validating.
    - Methods:``test_step``, supporting to forward when testing.

    Args:
        backbone (dict): Backbone modules to extract feature.
        head (dict): Head to process feature.
        loss(dict): Loss function.
    """
    def __init__(self, backbone=None, head=None, loss=None):
        super().__init__()
        if backbone != None:
            self.backbone = builder.build_backbone(backbone)
            if hasattr(self.backbone, 'init_weights'):
                self.backbone.init_weights()
        else:
            self.backbone = None
        if head != None:
            self.head_name = head.name
            if head.name == 'IntVOS':
                head.update({'feature_extracter': self.backbone})
                self.head = builder.build_head(head)
            else:
                self.head = builder.build_head(head)
            if hasattr(self.head, 'init_weights'):
                self.head.init_weights()
        else:
            self.head = None
        if loss != None:
            self.loss = builder.build_loss(loss)
        else:
            self.loss = None

    def forward(self, data_batch, mode='infer', **kwargs):
        """
        1. Define how the model is going to run, from input to output.
        2. Console of train, valid, test or infer step
        3. Set mode='infer' is used for saving inference model, refer to tools/export_model.py
        """
        if mode == 'train':
            return self.train_step(data_batch, **kwargs)
        elif mode == 'valid':
            return self.val_step(data_batch, **kwargs)
        elif mode == 'test':
            return self.test_step(data_batch, **kwargs)
        elif mode == 'infer':
            return self.infer_step(data_batch, **kwargs)
        else:
            raise NotImplementedError

    @abstractmethod
    def train_step(self, data_batch, **kwargs):
        """Training step.
        """
        raise NotImplementedError

    @abstractmethod
    def val_step(self, data_batch, **kwargs):
        """Validating step.
        """
        raise NotImplementedError

    @abstractmethod
    def test_step(self, data_batch, **kwargs):
        """Test step.
        """
        raise NotImplementedError

    @abstractmethod
    def infer_step(self, data_batch, **kwargs):
        """Infer step.
        """
        raise NotImplementedError



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import paddle
import paddle.nn.initializer as init
import numpy as np
from scipy import special


def weight_init_(layer,
                 func,
                 weight_name=None,
                 bias_name=None,
                 bias_value=0.0,
                 **kwargs):
    """
    In-place params init function.
    Usage:
    .. code-block:: python

        import paddle
        import numpy as np

        data = np.ones([3, 4], dtype='float32')
        linear = paddle.nn.Linear(4, 4)
        input = paddle.to_tensor(data)
        print(linear.weight)
        linear(input)

        weight_init_(linear, 'Normal', 'fc_w0', 'fc_b0', std=0.01, mean=0.1)
        print(linear.weight)
    """

    if hasattr(layer, 'weight') and layer.weight is not None:
        getattr(init, func)(**kwargs)(layer.weight)
        if weight_name is not None:
            # override weight name
            layer.weight.name = weight_name

    if hasattr(layer, 'bias') and layer.bias is not None:
        init.Constant(bias_value)(layer.bias)
        if bias_name is not None:
            # override bias name
            layer.bias.name = bias_name


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        print("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
              "The distribution of values may be incorrect.")

    with paddle.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to [2l-1, 2u-1].
        tmp = np.random.uniform(2 * l - 1, 2 * u - 1,
                                size=list(tensor.shape)).astype(np.float32)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tmp = special.erfinv(tmp)

        # Transform to proper mean, std
        tmp *= (std * math.sqrt(2.0))
        tmp += mean

        # Clamp to ensure it's in the proper range
        tmp = np.clip(tmp, a, b)
        tensor.set_value(paddle.to_tensor(tmp))

        return tensor


def _calculate_fan_in_and_fan_out(tensor):
    dimensions = tensor.dim()
    if dimensions < 2:
        raise ValueError(
            "Fan in and fan out can not be computed for tensor with fewer than 2 dimensions"
        )

    num_input_fmaps = tensor.shape[1]
    num_output_fmaps = tensor.shape[0]
    receptive_field_size = 1
    if tensor.dim() > 2:
        receptive_field_size = tensor[0][0].numel()
    fan_in = num_input_fmaps * receptive_field_size
    fan_out = num_output_fmaps * receptive_field_size

    return fan_in, fan_out


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def kaiming_normal_(tensor, a=0., mode='fan_in', nonlinearity='leaky_relu'):
    def _calculate_correct_fan(tensor, mode):
        mode = mode.lower()
        valid_modes = ['fan_in', 'fan_out']
        if mode not in valid_modes:
            raise ValueError(
                "Mode {} not supported, please use one of {}".format(
                    mode, valid_modes))

        fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
        return fan_in if mode == 'fan_in' else fan_out

    def calculate_gain(nonlinearity, param=None):
        linear_fns = [
            'linear', 'conv1d', 'conv2d', 'conv3d', 'conv_transpose1d',
            'conv_transpose2d', 'conv_transpose3d'
        ]
        if nonlinearity in linear_fns or nonlinearity == 'sigmoid':
            return 1
        elif nonlinearity == 'tanh':
            return 5.0 / 3
        elif nonlinearity == 'relu':
            return math.sqrt(2.0)
        elif nonlinearity == 'leaky_relu':
            if param is None:
                negative_slope = 0.01
            elif not isinstance(param, bool) and isinstance(
                    param, int) or isinstance(param, float):
                negative_slope = param
            else:
                raise ValueError(
                    "negative_slope {} not a valid number".format(param))
            return math.sqrt(2.0 / (1 + negative_slope**2))
        else:
            raise ValueError(
                "Unsupported nonlinearity {}".format(nonlinearity))

    fan = _calculate_correct_fan(tensor, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    with paddle.no_grad():
        paddle.nn.initializer.Normal(0, std)(tensor)
        return tensor



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools

import paddle
import paddle.distributed as dist

def get_dist_info():
    world_size = dist.get_world_size()
    rank = dist.get_rank()
    return rank, world_size

def main_only(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        rank, _ = get_dist_info()
        if rank == 0:
            return func(*args, **kwargs)
    return wrapper



import paddle.nn as nn
# from reprod_log.utils import paddle2np

from EIVideo.paddlevideo.utils.manet_utils import fill_, zero_


class Bottleneck(nn.Layer):
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 dilation=1,
                 downsample=None,
                 BatchNorm=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2D(inplanes, planes, kernel_size=1, bias_attr=False)
        self.bn1 = BatchNorm(planes)
        self.conv2 = nn.Conv2D(planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               dilation=dilation,
                               padding=dilation,
                               bias_attr=False)
        self.bn2 = BatchNorm(planes)
        self.conv3 = nn.Conv2D(planes,
                               planes * 4,
                               kernel_size=1,
                               bias_attr=False)
        self.bn3 = BatchNorm(planes * 4)
        self.relu = nn.ReLU()
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Layer):
    def __init__(self,
                 block,
                 layers,
                 output_stride,
                 BatchNorm,
                 pretrained=None):
        self.inplanes = 64
        super(ResNet, self).__init__()
        blocks = [1, 2, 4]
        if output_stride == 16:
            strides = [1, 2, 2, 1]
            dilations = [1, 1, 1, 2]
        elif output_stride == 8:
            strides = [1, 2, 1, 1]
            dilations = [1, 1, 2, 4]
        else:
            raise NotImplementedError

        # Modules
        self.conv1 = nn.Conv2D(3,
                               64,
                               kernel_size=7,
                               stride=2,
                               padding=3,
                               bias_attr=False)
        self.bn1 = BatchNorm(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2D(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block,
                                       64,
                                       layers[0],
                                       stride=strides[0],
                                       dilation=dilations[0],
                                       BatchNorm=BatchNorm)
        self.layer2 = self._make_layer(block,
                                       128,
                                       layers[1],
                                       stride=strides[1],
                                       dilation=dilations[1],
                                       BatchNorm=BatchNorm)
        self.layer3 = self._make_layer(block,
                                       256,
                                       layers[2],
                                       stride=strides[2],
                                       dilation=dilations[2],
                                       BatchNorm=BatchNorm)
        self.layer4 = self._make_MG_unit(block,
                                         512,
                                         blocks=blocks,
                                         stride=strides[3],
                                         dilation=dilations[3],
                                         BatchNorm=BatchNorm)
        self.init_weight()



    def _make_layer(self,
                    block,
                    planes,
                    blocks,
                    stride=1,
                    dilation=1,
                    BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2D(self.inplanes,
                          planes * block.expansion,
                          kernel_size=1,
                          stride=stride,
                          bias_attr=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(self.inplanes, planes, stride, dilation, downsample,
                  BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(
                block(self.inplanes,
                      planes,
                      dilation=dilation,
                      BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def _make_MG_unit(self,
                      block,
                      planes,
                      blocks,
                      stride=1,
                      dilation=1,
                      BatchNorm=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2D(self.inplanes,
                          planes * block.expansion,
                          kernel_size=1,
                          stride=stride,
                          bias_attr=False),
                BatchNorm(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(self.inplanes,
                  planes,
                  stride,
                  dilation=blocks[0] * dilation,
                  downsample=downsample,
                  BatchNorm=BatchNorm))
        self.inplanes = planes * block.expansion
        for i in range(1, len(blocks)):
            layers.append(
                block(self.inplanes,
                      planes,
                      stride=1,
                      dilation=blocks[i] * dilation,
                      BatchNorm=BatchNorm))

        return nn.Sequential(*layers)

    def forward(self, input):
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        low_level_feat = x
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        return x, low_level_feat

    def init_weight(self):
        for m in self.sublayers():
            if isinstance(m, nn.Conv2D):
                n = m._kernel_size[0] * m._kernel_size[1] * m._out_channels
                fill_(m.weight, 1)
            elif isinstance(m, nn.BatchNorm2D):
                fill_(m.weight, 1)
                zero_(m.bias)
        return self.sublayers()




def ResNet101(output_stride, BatchNorm, pretrained=None):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3],
                   output_stride,
                   BatchNorm,
                   pretrained=pretrained)
    return model


def build_backbone(output_stride, BatchNorm, pretrained):
    return ResNet101(output_stride, BatchNorm, pretrained)


if __name__ == "__main__":
    import paddle

    model = ResNet101(BatchNorm=nn.BatchNorm2D,
                      pretrained=True,
                      output_stride=8)
    input = paddle.rand([1, 3, 512, 512])
    output, low_level_feat = model(input)
    print(output.shape)
    print(low_level_feat.shape)
    import json

    with open('output.txt', 'w') as f:
        json.dump(output.tolist(), f)
    with open('low_level_feat.txt', 'w') as f:
        json.dump(low_level_feat.tolist(), f)



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import signal
import os
import paddle
from paddle.fluid.dataloader import BatchSampler
from paddle.io import DataLoader, DistributedBatchSampler

from .pipelines.compose import Compose
from .registry import DATASETS, PIPELINES, DATALOADERS, BATCH_SAMPLERS, SAMPLERS
from ..utils import get_logger
from ..utils.build_utils import build
import numpy as np

logger = get_logger("paddlevideo")


def build_pipeline(cfg):
    """Build pipeline.
    Args:
        cfg (dict): root config dict.
    """
    if cfg == None:
        return
    return Compose(cfg)


def build_dataset(cfg):
    """Build dataset.
    Args:
        cfg (dict): root config dict.

    Returns:
        dataset: dataset.
    """
    # XXX: ugly code here!
    cfg_dataset, cfg_pipeline = cfg
    cfg_dataset.pipeline = build_pipeline(cfg_pipeline)
    dataset = build(cfg_dataset, DATASETS, key="format")
    return dataset


def build_sampler(cfg):
    """Build batch_sampler.
    Args:
        cfg (dict): root config dict.

    Returns:
        batch_sampler: batch_sampler.
    """
    sampler = build(cfg, SAMPLERS)
    return sampler


def build_batch_pipeline(cfg):
    batch_pipeline = build(cfg, PIPELINES)
    return batch_pipeline


def build_custom_dataloader(cfg):
    custom_dataloader = build(cfg, DATALOADERS, key='dataloader')
    return custom_dataloader


def build_dataloader(dataset,
                     batch_size,
                     num_workers,
                     places=None,
                     shuffle=True,
                     drop_last=True,
                     multigrid=False,
                     collate_fn_cfg=None,
                     **kwargs):
    """Build Paddle Dataloader.

    XXX explain how the batch_sampler work!

    Args:
        dataset (paddle.dataset): A PaddlePaddle dataset object.
        batch_size (int): batch size on single card.
        num_worker (int): num_worker
        shuffle(bool): whether to shuffle the data at every epoch.
    """

    if not kwargs.get('sampler'):
        batch_sampler = DistributedBatchSampler(dataset,
                                                    batch_size=batch_size,
                                                    shuffle=shuffle,
                                                    drop_last=drop_last)
    else:
        sampler = build_sampler(kwargs['sampler'])
        batch_sampler = BatchSampler(dataset,
                                     sampler=sampler,
                                     batch_size=batch_size,
                                     shuffle=shuffle,
                                     drop_last=drop_last)
    kwargs.update({'batch_sampler': batch_sampler})

    # NOTE(shipping): when switch the mix operator on, such as: mixup, cutmix.

    # batch like: [[img, label, attibute, ...], [imgs, label, attribute, ...], ...] will recollate to:
    # [[img, img, ...], [label, label, ...], [attribute, attribute, ...], ...] as using numpy.transpose.

    def mix_collate_fn(batch):
        pipeline = build_batch_pipeline(collate_fn_cfg)
        batch = pipeline(batch)
        slots = []
        for items in batch:
            for i, item in enumerate(items):
                if len(slots) < len(items):
                    slots.append([item])
                else:
                    slots[i].append(item)
        return [np.stack(slot, axis=0) for slot in slots]

    # if collate_fn_cfg is not None:
    # ugly code here. collate_fn is mix op config
    #    collate_fn = mix_collate_fn(collate_fn_cfg)

    data_loader = DataLoader(
        dataset,
        places=places,
        num_workers=num_workers,
        collate_fn=mix_collate_fn if collate_fn_cfg is not None else None,
        **kwargs)

    return data_loader


def term_mp(sig_num, frame):
    """ kill all child processes
    """
    pid = os.getpid()
    pgid = os.getpgid(os.getpid())
    logger.info("main proc {} exit, kill process group " "{}".format(pid, pgid))
    os.killpg(pgid, signal.SIGKILL)
    return


signal.signal(signal.SIGINT, term_mp)
signal.signal(signal.SIGTERM, term_mp)



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def build(cfg, registry, key='name'):
    """Build a module from config dict.
    Args:
        cfg (dict): Config dict. It should at least contain the key.
        registry (XXX): The registry to search the type from.
        key (str): the key.
    Returns:
        obj: The constructed object.
    """

    assert isinstance(cfg, dict) and key in cfg

    cfg_copy = cfg.copy()
    obj_type = cfg_copy.pop(key)

    obj_cls = registry.get(obj_type)
    if obj_cls is None:
        raise KeyError('{} is not in the {} registry'.format(
                obj_type, registry.name))
    return obj_cls(**cfg_copy)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import datetime

from paddle.distributed import ParallelEnv



Color = {
    'RED': '\033[31m',
    'HEADER': '\033[35m',  # deep purple
    'PURPLE': '\033[95m',  # purple
    'OKBLUE': '\033[94m',
    'OKGREEN': '\033[92m',
    'WARNING': '\033[93m',
    'FAIL': '\033[91m',
    'ENDC': '\033[0m'
}


def coloring(message, color="OKGREEN"):
    assert color in Color.keys()
    if os.environ.get('COLORING', True):
        return Color[color] + str(message) + Color["ENDC"]
    else:
        return message


logger_initialized = []


def setup_logger(output=None, name="paddlevideo", level="INFO"):
    """
    Initialize the paddlevideo logger and set its verbosity level to "INFO".
    Args:
        output (str): a file name or a directory to save log. If None, will not save log file.
            If ends with ".txt" or ".log", assumed to be a file name.
            Otherwise, logs will be saved to `output/log.txt`.
        name (str): the root module name of this logger
    Returns:
        logging.Logger: a logger
    """
    def time_zone(sec, fmt):
        real_time = datetime.datetime.now()
        return real_time.timetuple()
    logging.Formatter.converter = time_zone

    logger = logging.getLogger(name)
    if level == "INFO":
        logger.setLevel(logging.INFO)
    elif level=="DEBUG":
        logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if level == "DEBUG":
        plain_formatter = logging.Formatter(
            "[%(asctime)s] %(name)s %(levelname)s: %(message)s",
            datefmt="%m/%d %H:%M:%S")
    else:
        plain_formatter = logging.Formatter(
            "[%(asctime)s] %(message)s",
            datefmt="%m/%d %H:%M:%S")
    # stdout logging: master only
    local_rank = ParallelEnv().local_rank
    if local_rank == 0:
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = plain_formatter
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    # file logging: all workers
    if output is not None:
        if output.endswith(".txt") or output.endswith(".log"):
            filename = output
        else:
            filename = os.path.join(output, ".log.txt")
        if local_rank > 0:
            filename = filename + ".rank{}".format(local_rank)

        # PathManager.mkdirs(os.path.dirname(filename))
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        # fh = logging.StreamHandler(_cached_log_stream(filename)
        fh = logging.FileHandler(filename, mode='a')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)
    logger_initialized.append(name)
    return logger


def get_logger(name, output=None):
    logger = logging.getLogger(name)
    if name in logger_initialized:
        return logger

    return setup_logger(name=name, output=name)



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Registry(object):
    """
    The registry that provides name -> object mapping, to support third-party users' custom modules.

    To register an object:

    .. code-block:: python

        BACKBONES = Registry('backbone')
        @BACKBONES.register()
        class ResNet:
            pass
    Or:
    .. code-block:: python

        BACKBONES = Registry('backbone')
        class ResNet:
            pass
        BACKBONES.register(ResNet)

    Usage: To build a module.

    .. code-block:: python
        backbone_name = "ResNet"
        b = BACKBONES.get(backbone_name)()

    """
    def __init__(self, name):
        """
        Args:
            name (str): the name of this registry
        """
        self._name = name
        self._obj_map = {}

    def __contains__(self, key):
        return self._obj_map.get(key) is not None

    def _do_register(self, name, obj):
        assert (
            name not in self._obj_map
        ), "An object named '{}' was already registered in '{}' registry!".format(
            name, self._name)
        self._obj_map[name] = obj

    def register(self, obj=None, name=None):
        """
        Register the given object under the the name `obj.__name__`.
        Can be used as either a decorator or not. See docstring of this class for usage.
        """
        if obj is None:
            # used as a decorator
            def deco(func_or_class, name=name):
                if name is None:
                    name = func_or_class.__name__
                self._do_register(name, func_or_class)
                return func_or_class

            return deco

        # used as a function call
        if name is None:
            name = obj.__name__
        self._do_register(name, obj)

    def get(self, name):
        """Get the registry record.

        Args:
            name (str): The class name.

        Returns:
            ret: The class.
        """
        ret = self._obj_map.get(name)
        if ret is None:
            raise KeyError(
                "No object named '{}' found in '{}' registry!".format(
                    name, self._name))

        return ret



import paddle
import paddle.nn as nn
import paddle.nn.functional as F
from EIVideo.paddlevideo.utils.manet_utils import kaiming_normal_


class Decoder(nn.Layer):
    def __init__(self, num_classes, backbone, BatchNorm):
        super(Decoder, self).__init__()
        if backbone == 'resnet' or backbone == 'drn' or backbone == 'resnet_edge':
            low_level_inplanes = 256
        elif backbone == 'xception':
            low_level_inplanes = 128
        elif backbone == 'mobilenet':
            low_level_inplanes = 24
        else:
            raise NotImplementedError

        self.conv1 = nn.Conv2D(low_level_inplanes, 48, 1, bias_attr=False)
        self.bn1 = BatchNorm(48)
        self.relu = nn.ReLU(True)
        self.last_conv = nn.Sequential(
            nn.Conv2D(304,
                      256,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias_attr=False), BatchNorm(256), nn.ReLU(True),
            nn.Sequential(),
            nn.Conv2D(256,
                      256,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias_attr=False), BatchNorm(256), nn.ReLU(True),
            nn.Sequential())
        self._init_weight()

    def forward(self, x, low_level_feat):
        low_level_feat = self.conv1(low_level_feat)
        low_level_feat = self.bn1(low_level_feat)
        low_level_feat = self.relu(low_level_feat)

        x = F.interpolate(x,
                          size=low_level_feat.shape[2:],
                          mode='bilinear',
                          align_corners=True)
        x = paddle.concat((x, low_level_feat), axis=1)
        x = self.last_conv(x)

        return x

    def _init_weight(self):
        for m in self.sublayers():
            if isinstance(m, nn.Conv2D):
                kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2D):
                from EIVideo.paddlevideo.utils.manet_utils import fill_
                fill_(m.weight, 1)
                from EIVideo.paddlevideo.utils.manet_utils import zero_
                zero_(m.bias)


def build_decoder(num_classes, backbone, BatchNorm):
    return Decoder(num_classes, backbone, BatchNorm)



# Copyright (c) 2020  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from ..registry import PIPELINES
import traceback
from ...utils import build
from ...utils import get_logger


@PIPELINES.register()
class Compose(object):
    """
    Composes several pipelines(include decode func, sample func, and transforms) together.

    Note: To deal with ```list``` type cfg temporaray, like:

        transform:
            - Crop: # A list
                attribute: 10
            - Resize: # A list
                attribute: 20

    every key of list will pass as the key name to build a module.
    XXX: will be improved in the future.

    Args:
        pipelines (list): List of transforms to compose.
    Returns:
        A compose object which is callable, __call__ for this Compose
        object will call each given :attr:`transforms` sequencely.
    """
    def __init__(self, pipelines):
        #assert isinstance(pipelines, Sequence)
        self.pipelines = []
        for p in pipelines.values():
            if isinstance(p, dict):
                p = build(p, PIPELINES)
                self.pipelines.append(p)
            elif isinstance(p, list):
                for t in p:
                    #XXX: to deal with old format cfg, ugly code here!
                    temp_dict = dict(name=list(t.keys())[0])
                    for all_sub_t in t.values():
                        if all_sub_t is not None:
                            temp_dict.update(all_sub_t) 
      
                    t = build(temp_dict, PIPELINES)
                    self.pipelines.append(t)
            elif callable(p):
                self.pipelines.append(p)
            else:
                raise TypeError(f'pipelines must be callable or a dict,'
                                f'but got {type(p)}')
    def __call__(self, data):
        for p in self.pipelines:
            try:
                data = p(data)
            except Exception as e:
                stack_info = traceback.format_exc()
                logger = get_logger("paddlevideo")
                logger.info("fail to perform transform [{}] with error: "
                      "{} and stack:\n{}".format(p, e, str(stack_info)))
                raise e
        return data



import os
import random
import cv2
import numpy as np
import paddle
from PIL import Image
from davisinteractive.utils.operations import bresenham

from ..registry import PIPELINES

cv2.setNumThreads(0)
NEW_BRANCH = True


@PIPELINES.register()
class RandomScale_manet(object):
    """Randomly resize the image and the ground truth to specified scales.
    Args:
        scales (list): the list of scales
    """
    def __init__(self, scales=[0.75, 1, 1.25]):
        self.scales = scales

    def __call__(self, sample):

        # Fixed range of scales
        sc = self.scales[random.randint(0, len(self.scales) - 1)]

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'img1' or elem == 'img2' or elem == 'ref_img':
                flagval = cv2.INTER_CUBIC
            else:
                flagval = cv2.INTER_NEAREST

            tmp = cv2.resize(tmp, None, fx=sc, fy=sc, interpolation=flagval)

            sample[elem] = tmp

        return sample


@PIPELINES.register()
class Resize_manet(object):
    """Rescale the image in a results to a given size.

    Args:
        output_size (tuple or int): Desired output size. If tuple, output is
            matched to output_size. If int, smaller of image edges is matched
            to output_size keeping aspect ratio the same.
    """
    def __init__(self, output_size):
        assert isinstance(output_size, (int, list))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size)
        else:
            self.output_size = output_size

    #        self.seg_interpolation = cv2.INTER_CUBIC if is_continuous else cv2.INTER_NEAREST
    #        self.fix = fix

    def __call__(self, results):
        img1 = results['img1']
        h, w = img1.shape[:2]
        if self.output_size == (h, w):
            return results

        else:
            new_h, new_w = self.output_size
        new_h, new_w = int(new_h), int(new_w)
        for elem in results.keys():
            if 'meta' in elem:
                continue
            tmp = results[elem]
            if elem == 'img1' or elem == 'img2' or elem == 'ref_img':
                flagval = cv2.INTER_CUBIC
            else:
                flagval = cv2.INTER_NEAREST

            tmp = cv2.resize(tmp, dsize=(new_w, new_h), interpolation=flagval)
            results[elem] = tmp
        return results


@PIPELINES.register()
class RandomCrop_manet(object):
    """Crop randomly the image in a results.

    Args:
        output_size (tuple or int): Desired output size. If int, square crop
            is made.
    """
    def __init__(self, output_size, step=None):
        assert isinstance(output_size, (int, list))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size)
        else:
            assert len(output_size) == 2
            self.output_size = output_size
        self.step = step

    def __call__(self, results):

        image = results['img1']
        h, w = image.shape[:2]
        new_h, new_w = self.output_size

        new_h = h if new_h >= h else new_h
        new_w = w if new_w >= w else new_w
        is_contain_obj = False

        #        while (not is_contain_obj) and (step < 5):
        if self.step is None:
            while not is_contain_obj:
                #                step += 1
                top = np.random.randint(0, h - new_h + 1)
                left = np.random.randint(0, w - new_w + 1)
                ref_scribble_label = results['ref_scribble_label']
                new_ref_scribble_label = ref_scribble_label[top:top + new_h,
                                                            left:left + new_w]
                if len(np.unique(new_ref_scribble_label)) == 1:
                    continue
                else:

                    for elem in results.keys():
                        if 'meta' in elem:
                            continue

                        tmp = results[elem]
                        tmp = tmp[top:top + new_h, left:left + new_w]
                        results[elem] = tmp
                    break
        else:
            st = 0
            while not is_contain_obj and st < self.step:
                st += 1
                top = np.random.randint(0, h - new_h + 1)
                left = np.random.randint(0, w - new_w + 1)
                ref_scribble_label = results['ref_scribble_label']
                new_ref_scribble_label = ref_scribble_label[top:top + new_h,
                                                            left:left + new_w]
                if len(np.unique(
                        new_ref_scribble_label)) == 1 or st < self.step - 1:
                    continue
                else:

                    for elem in results.keys():
                        if 'meta' in elem:
                            continue

                        tmp = results[elem]
                        tmp = tmp[top:top + new_h, left:left + new_w]
                        results[elem] = tmp
                    break

        return results


@PIPELINES.register()
class RandomHorizontalFlip_manet(object):
    """Horizontally flip the given image and ground truth randomly with a probability of 0.5."""
    def __init__(self, prob):
        self.p = prob

    def __call__(self, results):

        if random.random() < self.p:
            for elem in results.keys():
                if 'meta' in elem:
                    continue
                tmp = results[elem]
                tmp = cv2.flip(tmp, flipCode=1)
                results[elem] = tmp

        return results


@PIPELINES.register()
class ToTensor_manet(object):
    """Convert ndarrays in results to Tensors."""
    def __call__(self, results):

        for elem in results.keys():
            if 'meta' in elem:
                continue
            tmp = results[elem]

            if tmp.ndim == 2:
                tmp = tmp[:, :, np.newaxis]
            else:
                tmp = tmp / 255.
                tmp -= (0.485, 0.456, 0.406)
                tmp /= (0.229, 0.224, 0.225)
            tmp = tmp.transpose([2, 0, 1])
            results[elem] = paddle.to_tensor(tmp)
        return results


def gt_from_scribble(scr, dilation=11, nocare_area=21):
    # Compute foreground
    if scr.max() == 1:
        kernel_fg = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,
                                              (dilation, dilation))
        fg = cv2.dilate(scr.astype(np.uint8),
                        kernel=kernel_fg).astype(scr.dtype)
    else:
        fg = scr

    # Compute nocare area
    if nocare_area is None:
        nocare = None
    else:
        kernel_nc = cv2.getStructuringElement(cv2.MORPH_ELLIPSE,
                                              (nocare_area, nocare_area))
        nocare = cv2.dilate(fg, kernel=kernel_nc) - fg

    return fg, nocare



from __future__ import absolute_import

import json
import math
import os
import pickle
import warnings

import numpy
import numpy as np
from numpy import inf
from paddle import Tensor, concat, reshape, nn
import paddle

from typing import Union, Iterable

# from reprod_log.compare import compute_diff
# from reprod_log.utils import check_print_diff, np2torch, np2paddle, torch2np, paddle2np
def show_memory(unit='KB', threshold=1):
    '''查看变量占用内存情况

    :param unit: 显示的单位，可为`B`,`KB`,`MB`,`GB`
    :param threshold: 仅显示内存数值大于等于threshold的变量
    '''
    from sys import getsizeof
    scale = {'B': 1, 'KB': 1024, 'MB': 1048576, 'GB': 1073741824}[unit]
    for i in list(globals().keys()):
        memory = eval("getsizeof({})".format(i)) // scale
        if memory >= threshold:
            print(i, memory)





_tensor_or_tensors = Union[paddle.Tensor, Iterable[paddle.Tensor]]
_palette = [
    0, 0, 0, 128, 0, 0, 0, 128, 0, 128, 128, 0, 0, 0, 128, 128, 0, 128, 0, 128,
    128, 128, 128, 128, 64, 0, 0, 191, 0, 0, 64, 128, 0, 191, 128, 0, 64, 0,
    128, 191, 0, 128, 64, 128, 128, 191, 128, 128, 0, 64, 0, 128, 64, 0, 0,
    191, 0, 128, 191, 0, 0, 64, 128, 128, 64, 128, 22, 22, 22, 23, 23, 23, 24,
    24, 24, 25, 25, 25, 26, 26, 26, 27, 27, 27, 28, 28, 28, 29, 29, 29, 30, 30,
    30, 31, 31, 31, 32, 32, 32, 33, 33, 33, 34, 34, 34, 35, 35, 35, 36, 36, 36,
    37, 37, 37, 38, 38, 38, 39, 39, 39, 40, 40, 40, 41, 41, 41, 42, 42, 42, 43,
    43, 43, 44, 44, 44, 45, 45, 45, 46, 46, 46, 47, 47, 47, 48, 48, 48, 49, 49,
    49, 50, 50, 50, 51, 51, 51, 52, 52, 52, 53, 53, 53, 54, 54, 54, 55, 55, 55,
    56, 56, 56, 57, 57, 57, 58, 58, 58, 59, 59, 59, 60, 60, 60, 61, 61, 61, 62,
    62, 62, 63, 63, 63, 64, 64, 64, 65, 65, 65, 66, 66, 66, 67, 67, 67, 68, 68,
    68, 69, 69, 69, 70, 70, 70, 71, 71, 71, 72, 72, 72, 73, 73, 73, 74, 74, 74,
    75, 75, 75, 76, 76, 76, 77, 77, 77, 78, 78, 78, 79, 79, 79, 80, 80, 80, 81,
    81, 81, 82, 82, 82, 83, 83, 83, 84, 84, 84, 85, 85, 85, 86, 86, 86, 87, 87,
    87, 88, 88, 88, 89, 89, 89, 90, 90, 90, 91, 91, 91, 92, 92, 92, 93, 93, 93,
    94, 94, 94, 95, 95, 95, 96, 96, 96, 97, 97, 97, 98, 98, 98, 99, 99, 99,
    100, 100, 100, 101, 101, 101, 102, 102, 102, 103, 103, 103, 104, 104, 104,
    105, 105, 105, 106, 106, 106, 107, 107, 107, 108, 108, 108, 109, 109, 109,
    110, 110, 110, 111, 111, 111, 112, 112, 112, 113, 113, 113, 114, 114, 114,
    115, 115, 115, 116, 116, 116, 117, 117, 117, 118, 118, 118, 119, 119, 119,
    120, 120, 120, 121, 121, 121, 122, 122, 122, 123, 123, 123, 124, 124, 124,
    125, 125, 125, 126, 126, 126, 127, 127, 127, 128, 128, 128, 129, 129, 129,
    130, 130, 130, 131, 131, 131, 132, 132, 132, 133, 133, 133, 134, 134, 134,
    135, 135, 135, 136, 136, 136, 137, 137, 137, 138, 138, 138, 139, 139, 139,
    140, 140, 140, 141, 141, 141, 142, 142, 142, 143, 143, 143, 144, 144, 144,
    145, 145, 145, 146, 146, 146, 147, 147, 147, 148, 148, 148, 149, 149, 149,
    150, 150, 150, 151, 151, 151, 152, 152, 152, 153, 153, 153, 154, 154, 154,
    155, 155, 155, 156, 156, 156, 157, 157, 157, 158, 158, 158, 159, 159, 159,
    160, 160, 160, 161, 161, 161, 162, 162, 162, 163, 163, 163, 164, 164, 164,
    165, 165, 165, 166, 166, 166, 167, 167, 167, 168, 168, 168, 169, 169, 169,
    170, 170, 170, 171, 171, 171, 172, 172, 172, 173, 173, 173, 174, 174, 174,
    175, 175, 175, 176, 176, 176, 177, 177, 177, 178, 178, 178, 179, 179, 179,
    180, 180, 180, 181, 181, 181, 182, 182, 182, 183, 183, 183, 184, 184, 184,
    185, 185, 185, 186, 186, 186, 187, 187, 187, 188, 188, 188, 189, 189, 189,
    190, 190, 190, 191, 191, 191, 192, 192, 192, 193, 193, 193, 194, 194, 194,
    195, 195, 195, 196, 196, 196, 197, 197, 197, 198, 198, 198, 199, 199, 199,
    200, 200, 200, 201, 201, 201, 202, 202, 202, 203, 203, 203, 204, 204, 204,
    205, 205, 205, 206, 206, 206, 207, 207, 207, 208, 208, 208, 209, 209, 209,
    210, 210, 210, 211, 211, 211, 212, 212, 212, 213, 213, 213, 214, 214, 214,
    215, 215, 215, 216, 216, 216, 217, 217, 217, 218, 218, 218, 219, 219, 219,
    220, 220, 220, 221, 221, 221, 222, 222, 222, 223, 223, 223, 224, 224, 224,
    225, 225, 225, 226, 226, 226, 227, 227, 227, 228, 228, 228, 229, 229, 229,
    230, 230, 230, 231, 231, 231, 232, 232, 232, 233, 233, 233, 234, 234, 234,
    235, 235, 235, 236, 236, 236, 237, 237, 237, 238, 238, 238, 239, 239, 239,
    240, 240, 240, 241, 241, 241, 242, 242, 242, 243, 243, 243, 244, 244, 244,
    245, 245, 245, 246, 246, 246, 247, 247, 247, 248, 248, 248, 249, 249, 249,
    250, 250, 250, 251, 251, 251, 252, 252, 252, 253, 253, 253, 254, 254, 254,
    255, 255, 255
]


import PIL
from PIL import Image
from paddle.vision.transforms import BaseTransform

import numpy as np
from scipy.ndimage import interpolation, binary_dilation
from skimage import transform
import paddle
import cv2
import random


####
def mask_damager(labels=None, p_black=0.2):
    scales = (0.8, 1.0, 1.2)
    kernel_size = random.randint(10, 15)
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    if random.random() < p_black:
        final_label = paddle.zeros_like(labels)
        final_label = final_label.squeeze().numpy()
    else:
        prot = random.randint(5, 15)
        nrot = random.randint(-15, -5)
        rots = [prot, nrot, 0]
        rot = rots[random.randint(0, 2)]

        sc = scales[random.randint(0, 2)]
        _, _, h, w = labels.shape
        tmp = labels.squeeze()

        tmp = tmp.unsqueeze(-1)
        tmp = tmp.numpy().astype(np.uint8)
        morph_p = random.random()
        if morph_p < 0.5:
            tmp = cv2.morphologyEx(tmp, cv2.MORPH_OPEN, kernel)
        else:
            tmp = cv2.morphologyEx(tmp, cv2.MORPH_CLOSE, kernel)

        tmp = tmp.astype(np.uint8)
        center = (w / 2, h / 2)
        M = cv2.getRotationMatrix2D(center, rot, sc)
        final_label = cv2.warpAffine(tmp, M, (w, h), cv2.INTER_NEAREST)

    return final_label


color_map = [
    [0, 0, 0],
    [255, 127, 0],
    [30, 144, 255],
    [186, 85, 211],
    [255, 105, 180],
    [192, 255, 62],
    [255, 105, 180],
    [50, 255, 255],
]

color_map_np = np.array(color_map)


def overlay_davis(image, mask, alpha=0.5):
    """ Overlay segmentation on top of RGB image. from davis official"""
    im_overlay = image.copy()
    colored_mask = color_map_np[mask]
    foreground = image * alpha + (1 - alpha) * colored_mask
    binary_mask = (mask > 0)
    # Compose image
    im_overlay[binary_mask] = foreground[binary_mask]
    countours = binary_dilation(binary_mask) ^ binary_mask
    im_overlay[countours, :] = 0
    return im_overlay.astype(image.dtype)


# TODO
def submit_masks(masks, images, inter_file_path):
    overlays = []
    save_result_path = os.path.join(inter_file_path, 'result')
    os.makedirs(save_result_path, exist_ok=True)
    for imgname, (mask, image) in enumerate(zip(masks, images)):
        overlay = overlay_davis(image, mask)
        overlays.append(overlay.tolist())
        overlay = Image.fromarray(overlay)
        imgname = str(imgname)
        while len(imgname) < 5:
            imgname = '0' + imgname
        overlay.save(os.path.join(save_result_path, imgname + '.png'))
    result = {'overlays': overlays}
    # result = {'masks': masks.tolist()}
    with open(os.path.join(save_result_path, 'masks.json'), 'w') as f:
        json.dump(result, f)


def load_video(path, min_side=None):
    frame_list = []
    cap = cv2.VideoCapture(path)
    while (cap.isOpened()):
        _, frame = cap.read()
        if frame is None:
            break
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        if min_side:
            h, w = frame.shape[:2]
            new_w = (w * min_side // min(w, h))
            new_h = (h * min_side // min(w, h))
            frame = cv2.resize(frame, (new_w, new_h),
                               interpolation=cv2.INTER_CUBIC)
            # .transpose([2, 0, 1])
        frame_list.append(frame)
    frames = np.stack(frame_list, axis=0)
    return frames


def get_scribbles():
    for i in range(8):
        with open(f'/home/lc/paddlevideo/data/bike-packing/lable/{i + 1}.json'
                  ) as f:
            scribbles = json.load(f)
            first_scribble = not i
            yield scribbles, first_scribble


def get_images(sequence='bike-packing'):
    img_path = os.path.join('/home/lc/paddlevideo/data', sequence.strip(),
                            'frame')
    img_files = os.listdir(img_path)
    img_files.sort()
    files = []
    for img in img_files:
        img_file = np.array(Image.open(os.path.join(img_path, img)))
        files.append(img_file)
    return np.array(files)


def rough_ROI(ref_scribble_labels):
    #### b*1*h*w
    dist = 20
    b, _, h, w = ref_scribble_labels.shape
    filter_ = paddle.zeros_like(ref_scribble_labels)
    to_fill = paddle.zeros_like(ref_scribble_labels)
    for i in range(b):
        no_background = (ref_scribble_labels[i] != -1)
        no_background = no_background.squeeze(0)

        no_b = no_background.nonzero()
        (h_min, w_min) = paddle.min(no_b, 0)
        (h_max, w_max) = paddle.max(no_b, 0)
        filter_[i, 0,
                max(h_min - dist, 0):min(h_max + dist, h - 1),
                max(w_min - dist, 0):min(w_max + dist, w - 1)] = 1

    final_scribble_labels = paddle.where(byte_(filter_), ref_scribble_labels,
                                         to_fill)
    return final_scribble_labels


import os.path as osp


def load(file_name, model, **cfg):
    if not osp.isfile(file_name):
        raise IOError(f'{file_name} not exist')
    try:
        state_dicts_ = paddle.load(file_name)['state_dict']
    except:
        state_dicts_ = paddle.load(file_name)
    state_dicts = {}
    for k in model.keys():
        if 'num_batches_tracked' not in k:
            if ('head.' + k) not in state_dicts_.keys():
                if k not in state_dicts_.keys():
                    print(f'model -----{k} -------is not in pretrained')
                else:
                    state_dicts[k] = state_dicts_[k]
            else:
                state_dicts[k] = state_dicts_['head.' + k]
    write_dict(state_dicts, 'state_dicts.txt', **cfg)
    write_dict(model, 'model.txt', **cfg)
    return state_dicts


#####
def write_dict(state_dict, file_name, **cfg):
    lines = []
    tot = 0
    for k, v in state_dict.items():
        # 目前只发现了torch和paddle模型参数命名的这三种不一致
        # 不一致1
        if 'num_batches_tracked' in k:
            tot += 1
            continue
        try:
            line = str(k) + '\t' + str(v.cpu().detach().numpy().shape) + '\n'
        except:
            line = str(k) + '\t' + str(v.shape) + '\n'
        lines.append(line)
    # with open(cfg.get("output_dir", f"./output/{file_name}"), 'w') as f:
    #     f.writelines(lines)
    # print('%d num_batches_tracked skipped' % tot)


def damage_masks(labels, shift=True, scale=True, rotate=True):
    """
    Args:
    labels: numpy array (batch_size * 1 * h * w)
    """
    bs, _, h, w = labels.shape
    labels = labels.transpose([0, 2, 3, 1])
    labels = labels.numpy()
    final_label = []
    for i in range(bs):
        label = labels[i]
        damaged_label = damage_masks_np(label, shift, scale, rotate)
        final_label.append(damaged_label)
    final_label = np.array(final_label)
    final_label = paddle.to_tensor(final_label)
    final_label = final_label.transpose([0, 3, 1, 2])
    return final_label


def damage_masks_np(labels, shift=True, scale=True, rotate=True):
    """Performs the actual mask damaging in numpy.
    Args:
    labels: Int32 numpy array of shape (height, width, 1).
    shift: Boolean, whether to damage the masks by shifting.
    scale: Boolean, whether to damage the masks by scaling.
    rotate: Boolean, whether to damage the masks by rotation.
    dilate: Boolean, whether to damage the masks by dilation.
    Returns:
    The damaged version of labels.
    """
    unique_labels = np.unique(labels)
    unique_labels = np.setdiff1d(unique_labels, [0])
    # Shuffle to get random depth ordering when combining together.
    np.random.shuffle(unique_labels)
    damaged_labels = np.zeros_like(labels)
    for l in unique_labels:
        obj_mask = (labels == l)
        damaged_obj_mask = _damage_single_object_mask(obj_mask, shift, scale,
                                                      rotate)
        damaged_labels[damaged_obj_mask] = l
    return damaged_labels


def _damage_single_object_mask(mask, shift, scale, rotate):
    """Performs mask damaging in numpy for a single object.
    Args:
    mask: Boolean numpy array of shape(height, width, 1).
    shift: Boolean, whether to damage the masks by shifting.
    scale: Boolean, whether to damage the masks by scaling.
    rotate: Boolean, whether to damage the masks by rotation.
    dilate: Boolean, whether to damage the masks by dilation.
    Returns:
    The damaged version of mask.
    """
    if shift:
        mask = _shift_mask(mask)
    if scale:
        mask = _scale_mask(mask)
    if rotate:
        mask = _rotate_mask(mask)
    return mask


def _shift_mask(mask, max_shift_factor=0.05):
    """Damages a mask for a single object by randomly shifting it in numpy.
    Args:
    mask: Boolean numpy array of shape(height, width, 1).
    max_shift_factor: Float scalar, the maximum factor for random shifting.
    Returns:
    The shifted version of mask.
    """
    nzy, nzx, _ = mask.nonzero()
    h = nzy.max() - nzy.min()
    w = nzx.max() - nzx.min()
    size = np.sqrt(h * w)
    offset = np.random.uniform(-size * max_shift_factor,
                               size * max_shift_factor, 2)
    shifted_mask = interpolation.shift(np.squeeze(mask, axis=2),
                                       offset,
                                       order=0).astype('bool')[..., np.newaxis]
    return shifted_mask


def _scale_mask(mask, scale_amount=0.025):
    """Damages a mask for a single object by randomly scaling it in numpy.
    Args:
    mask: Boolean numpy array of shape(height, width, 1).
    scale_amount: Float scalar, the maximum factor for random scaling.
    Returns:
    The scaled version of mask.
    """
    nzy, nzx, _ = mask.nonzero()
    cy = 0.5 * (nzy.max() - nzy.min())
    cx = 0.5 * (nzx.max() - nzx.min())
    scale_factor = np.random.uniform(1.0 - scale_amount, 1.0 + scale_amount)
    shift = transform.SimilarityTransform(translation=[-cx, -cy])
    inv_shift = transform.SimilarityTransform(translation=[cx, cy])
    s = transform.SimilarityTransform(scale=[scale_factor, scale_factor])
    m = (shift + (s + inv_shift)).inverse
    scaled_mask = transform.warp(mask, m) > 0.5
    return scaled_mask


def _rotate_mask(mask, max_rot_degrees=3.0):
    """Damages a mask for a single object by randomly rotating it in numpy.
    Args:
    mask: Boolean numpy array of shape(height, width, 1).
    max_rot_degrees: Float scalar, the maximum number of degrees to rotate.
    Returns:
    The scaled version of mask.
    """
    cy = 0.5 * mask.shape[0]
    cx = 0.5 * mask.shape[1]
    rot_degrees = np.random.uniform(-max_rot_degrees, max_rot_degrees)
    shift = transform.SimilarityTransform(translation=[-cx, -cy])
    inv_shift = transform.SimilarityTransform(translation=[cx, cy])
    r = transform.SimilarityTransform(rotation=np.deg2rad(rot_degrees))
    m = (shift + (r + inv_shift)).inverse
    scaled_mask = transform.warp(mask, m) > 0.5
    return scaled_mask


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


import numpy as np


def label2colormap(label):
    m = label.astype(np.uint8)
    r, c = m.shape
    cmap = np.zeros((r, c, 3), dtype=np.uint8)
    cmap[:, :, 0] = (m & 1) << 7 | (m & 8) << 3 | (m & 64) >> 1
    cmap[:, :, 1] = (m & 2) << 6 | (m & 16) << 2 | (m & 128) >> 2
    cmap[:, :, 2] = (m & 4) << 5 | (m & 32) << 1
    return cmap


def torch2paddle(data):
    try:
        import torch
        if isinstance(data, dict):
            np_data = {}
            for k, v in data.items():
                np_data[k] = paddle.to_tensor(v.detach().numpy())
            return np_data
        else:
            return paddle.to_tensor(data.detach().numpy())
    except:
        pass


def fill_(tensor: Tensor, value):
    return tensor.set_value(paddle.full_like(tensor, value))


def zero_(tensor: Tensor):
    return tensor.set_value(paddle.zeros_like(tensor))


def float_(tensor: Tensor):
    return paddle.to_tensor(tensor, dtype='float32')


def long_(tensor: Tensor):
    return paddle.to_tensor(tensor, dtype='int64')


def int_(tensor: Tensor):
    return paddle.to_tensor(tensor, dtype='int32')


def byte_(tensor: Tensor):
    return paddle.to_tensor(tensor, dtype='bool')


class ToPILImage(BaseTransform):
    def __init__(self, mode=None, keys=None):
        super(ToPILImage, self).__init__(keys)

    def _apply_image(self, pic):
        """
        Args:
            pic (Tensor|np.ndarray): Image to be converted to PIL Image.
        Returns:
            PIL: Converted image.
        """
        if not (isinstance(pic, paddle.Tensor) or isinstance(pic, np.ndarray)):
            raise TypeError('pic should be Tensor or ndarray. Got {}.'.format(
                type(pic)))

        elif isinstance(pic, paddle.Tensor):
            if pic.ndimension() not in {2, 3}:
                raise ValueError(
                    'pic should be 2/3 dimensional. Got {} dimensions.'.format(
                        pic.ndimension()))

            elif pic.ndimension() == 2:
                # if 2D image, add channel dimension (CHW)
                pic = pic.unsqueeze(0)

        elif isinstance(pic, np.ndarray):
            if pic.ndim not in {2, 3}:
                raise ValueError(
                    'pic should be 2/3 dimensional. Got {} dimensions.'.format(
                        pic.ndim))

            elif pic.ndim == 2:
                # if 2D image, add channel dimension (HWC)
                pic = np.expand_dims(pic, 2)

        npimg = pic
        if isinstance(pic, paddle.Tensor) and "float" in str(
                pic.numpy().dtype) and self.mode != 'F':
            pic = pic.mul(255).byte()
        if isinstance(pic, paddle.Tensor):
            npimg = np.transpose(pic.numpy(), (1, 2, 0))

        if not isinstance(npimg, np.ndarray):
            raise TypeError(
                'Input pic must be a paddle.Tensor or NumPy ndarray, ' +
                'not {}'.format(type(npimg)))

        if npimg.shape[2] == 1:
            expected_mode = None
            npimg = npimg[:, :, 0]
            if npimg.dtype == np.uint8:
                expected_mode = 'L'
            elif npimg.dtype == np.int16:
                expected_mode = 'I;16'
            elif npimg.dtype == np.int32:
                expected_mode = 'I'
            elif npimg.dtype == np.float32:
                expected_mode = 'F'
            if self.mode is not None and self.mode != expected_mode:
                raise ValueError(
                    "Incorrect self.mode ({}) supplied for input type {}. Should be {}"
                    .format(self.mode, np.dtype, expected_mode))
            self.mode = expected_mode

        elif npimg.shape[2] == 2:
            permitted_2_channel_modes = ['LA']
            if self.mode is not None and self.mode not in permitted_2_channel_modes:
                raise ValueError(
                    "Only self.modes {} are supported for 2D inputs".format(
                        permitted_2_channel_modes))

            if self.mode is None and npimg.dtype == np.uint8:
                self.mode = 'LA'

        elif npimg.shape[2] == 4:
            permitted_4_channel_modes = ['RGBA', 'CMYK', 'RGBX']
            if self.mode is not None and self.mode not in permitted_4_channel_modes:
                raise ValueError(
                    "Only self.modes {} are supported for 4D inputs".format(
                        permitted_4_channel_modes))

            if self.mode is None and npimg.dtype == np.uint8:
                self.mode = 'RGBA'
        else:
            permitted_3_channel_modes = ['RGB', 'YCbCr', 'HSV']
            if self.mode is not None and self.mode not in permitted_3_channel_modes:
                raise ValueError(
                    "Only self.modes {} are supported for 3D inputs".format(
                        permitted_3_channel_modes))
            if self.mode is None and npimg.dtype == np.uint8:
                self.mode = 'RGB'

        if self.mode is None:
            raise TypeError('Input type {} is not supported'.format(
                npimg.dtype))

        return Image.fromarray(npimg, mode=self.mode)


class Identity(nn.Layer):
    r"""A placeholder identity operator that is argument-insensitive.

    Args:
        args: any argument (unused)
        kwargs: any keyword argument (unused)
    """
    def __init__(self, *args, **kwargs):
        super(Identity, self).__init__()

    def forward(self, input):
        return input


def convert(data: dict, to, dtype=None):
    assert isinstance(data, dict)
    input = {}
    for k, v in data.items():

        if 'paddle' == to:
            if isinstance(v, np.ndarray):
                if dtype is not None:
                    input[k] = paddle.to_tensor(v.astype(dtype))
                else:
                    input[k] = paddle.to_tensor(v)
            else:
                input[k] = v
        elif 'torch' == to:
            try:
                import torch
                if isinstance(v, np.ndarray):
                    if dtype is not None:
                        input[k] = torch.tensor(v.astype(dtype))
                    else:
                        input[k] = torch.tensor(v)
                else:
                    input[k] = v
            except:
                pass
        else:
            if isinstance(v, np.ndarray):
                input[k] = v.astype(to)
            else:
                input[k] = v
    return input


def clip_grad_norm_(parameters: _tensor_or_tensors,
                    max_norm: float,
                    norm_type: float = 2.0,
                    error_if_nonfinite: bool = False) -> paddle.Tensor:
    r"""Clips gradient norm of an iterable of parameters.

    The norm is computed over all gradients together, as if they were
    concatenated into a single vector. Gradients are modified in-place.

    Args:
        parameters (Iterable[Tensor] or Tensor): an iterable of Tensors or a
            single Tensor that will have gradients normalized
        max_norm (float or int): max norm of the gradients
        norm_type (float or int): type of the used p-norm. Can be ``'inf'`` for
            infinity norm.
        error_if_nonfinite (bool): if True, an error is thrown if the total
            norm of the gradients from :attr:``parameters`` is ``nan``,
            ``inf``, or ``-inf``. Default: False (will switch to True in the future)

    Returns:
        Total norm of the parameters (viewed as a single vector).
    """
    import time
    if isinstance(parameters, paddle.Tensor):
        parameters = [parameters]
    parameters = [p for p in parameters if p.grad is not None]
    detached_grads = [p.grad.detach() for p in parameters]

    max_norm = float(max_norm)
    norm_type = float(norm_type)
    if len(parameters) == 0:
        return paddle.to_tensor(0.)
    # device = paddle.get_device()  # parameters[0].grad.device
    if norm_type == inf:
        norms = [p.abs().max() for p in parameters]
        total_norm = norms[0] if len(norms) == 1 else paddle.max(
            paddle.stack(norms))
    else:
        #         tik = time.time()
        total_norm = paddle.norm(
            paddle.stack([paddle.norm(g, norm_type) for g in detached_grads]),
            norm_type)
    #         total_norm = paddle.norm(paddle.stack([paddle.sqrt(paddle.sum(g*g)) for g in detached_grads]), norm_type)  # fixed.
    #         print(time.time() - tik)
    if error_if_nonfinite and paddle.logical_or(total_norm.isnan(),
                                                total_norm.isinf()):
        raise RuntimeError(
            f'The total norm of order {norm_type} for gradients from '
            '`parameters` is non-finite, so it cannot be clipped. To disable '
            'this error and scale the gradients by the non-finite norm anyway, '
            'set `error_if_nonfinite=False`')
    clip_coef = max_norm / (total_norm + 1e-6)
    # Note: multiplying by the clamped coef is redundant when the coef is clamped to 1, but doing so
    # avoids a `if clip_coef < 1:` conditional which can require a CPU <=> device synchronization
    # when the gradients do not reside in CPU memory.
    clip_coef_clamped = paddle.clip(clip_coef, max=1.0)
    for i, p in enumerate(parameters):
        #         p.set_value(paddle.multiply(p, clip_coef_clamped))
        p.grad.set_value(detached_grads[i] * clip_coef_clamped)  # fixed
    #         p.grad.detach().mul_(clip_coef_clamped
    return total_norm


# def max(a: paddle.Tensor, axis=0, keepdim=True):
#     """ndarray=numpy.array([[1, 2, 3, 4],
#            [4, 3, 2, 1],
#            [5, 6, 7, 8],
#            [8, 7, 6, 5]])
#     np.where(ndarray == np.max(ndarray))
#     (array([2, 3]), array([3, 0]))
#     ndarray[np.where(ndarray == np.max(ndarray))]
#     array([8, 8])
#     """
#     max_ = a.max(axis).unsqueeze(-1)
#     index = paddle.argmax(a, axis=axis, keepdim=keepdim)
#     max_ = max_.numpy()
#     index = index.numpy()
#     # index = paddle.argmax(a, axis=axis, keepdim=keepdim)[-1].flatten()
#     return max_, index


def gather(tmp: paddle.Tensor, ind: paddle.Tensor):
    shape = tmp.shape
    tmp = paddle.to_tensor(tmp)
    ind = paddle.to_tensor(ind)
    if len(shape) == 2:
        b = shape[0]
        return concat([
            reshape(paddle.gather(tmp[i, :], ind[i, :]), [1, -1])
            for i in range(b)
        ],
                      axis=0)
    elif len(shape) == 3:
        out = []
        for i in range(tmp.shape[0]):
            _ = paddle.index_sample(tmp[i], ind[i])
            out.append(_)
        return paddle.to_tensor(out)
    elif len(shape) == 4:
        b, c, d = shape[:3]
        return concat([
            reshape(
                concat([
                    reshape(
                        concat([
                            reshape(
                                paddle.gather(tmp[i, j, k, :],
                                              ind[i, j, k, :]), [1, -1])
                            for k in range(d)
                        ],
                               axis=0), [1, d, -1]) for j in range(c)
                ],
                       axis=0), [1, c, d, -1]) for i in range(b)
        ],
                      axis=0)
    else:
        pass


# These no_grad_* functions are necessary as wrappers around the parts of these
# functions that use `with torch.no_grad()`. The JIT doesn't support context
# managers, so these need to be implemented as builtins. Using these wrappers
# lets us keep those builtins small and re-usable.
def _no_grad_uniform_(tensor, a, b):
    with paddle.no_grad():
        tensor.set_value(paddle.uniform(tensor.shape, min=a, max=b))
        return tensor


def _no_grad_normal_(tensor, mean, std):
    with paddle.no_grad():
        tensor.set_value(paddle.normal(shape=tensor.shape, mean=mean, std=std))
        return tensor


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    from scipy import special

    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2)

    with paddle.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.set_value(
            paddle.uniform(tensor.shape, min=2 * l - 1, max=2 * u - 1))
        # tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        # tensor.erfinv_()  # paddle 无
        tensor.set_value(special.erfinv(tensor))

        # Transform to proper mean, std
        # tensor.mul_(std * math.sqrt(2.))
        tensor.set_value(tensor.multiply(paddle.to_tensor(std *
                                                          math.sqrt(2.))))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clip_(min=a, max=b)
        return tensor


def _no_grad_fill_(tensor, val):
    with paddle.no_grad():
        tensor.set_value(paddle.full_like(tensor, fill_value=val))
        return tensor


def _no_grad_zero_(tensor):
    with paddle.no_grad():
        tensor.set_value(paddle.zeros_like(tensor))
        return tensor


def calculate_gain(nonlinearity, param=None):
    r"""Return the recommended gain value for the given nonlinearity function.
    The values are as follows:

    ================= ====================================================
    nonlinearity      gain
    ================= ====================================================
    Linear / Identity :math:`1`
    Conv{1,2,3}D      :math:`1`
    Sigmoid           :math:`1`
    Tanh              :math:`\frac{5}{3}`
    ReLU              :math:`\sqrt{2}`
    Leaky Relu        :math:`\sqrt{\frac{2}{1 + \text{negative\_slope}^2}}`
    SELU              :math:`\frac{3}{4}`
    ================= ====================================================

    Args:
        nonlinearity: the non-linear function (`nn.functional` name)
        param: optional parameter for the non-linear function

    Examples:
        >>> gain = nn.init.calculate_gain('leaky_relu', 0.2)  # leaky_relu with negative_slope=0.2
    """
    linear_fns = [
        'linear', 'conv1d', 'conv2d', 'conv3d', 'conv_transpose1d',
        'conv_transpose2d', 'conv_transpose3d'
    ]
    if nonlinearity in linear_fns or nonlinearity == 'sigmoid':
        return 1
    elif nonlinearity == 'tanh':
        return 5.0 / 3
    elif nonlinearity == 'relu':
        return math.sqrt(2.0)
    elif nonlinearity == 'leaky_relu':
        if param is None:
            negative_slope = 0.01
        elif not isinstance(param, bool) and isinstance(
                param, int) or isinstance(param, float):
            # True/False are instances of int, hence check above
            negative_slope = param
        else:
            raise ValueError(
                "negative_slope {} not a valid number".format(param))
        return math.sqrt(2.0 / (1 + negative_slope**2))
    elif nonlinearity == 'selu':
        return 3.0 / 4  # Value found empirically (https://github.com/pytorch/pytorch/pull/50664)
    else:
        raise ValueError("Unsupported nonlinearity {}".format(nonlinearity))


def uniform_(tensor: Tensor, a: float = 0., b: float = 1.) -> Tensor:
    r"""Fills the input Tensor with values drawn from the uniform
    distribution :math:`\mathcal{U}(a, b)`.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        a: the lower bound of the uniform distribution
        b: the upper bound of the uniform distribution

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.uniform_(w)
    """
    return _no_grad_uniform_(tensor, a, b)


def normal_(tensor: Tensor, mean: float = 0., std: float = 1.) -> Tensor:
    r"""Fills the input Tensor with values drawn from the normal
    distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.normal_(w)
    """
    return _no_grad_normal_(tensor, mean, std)


def trunc_normal_(tensor: Tensor,
                  mean: float = 0.,
                  std: float = 1.,
                  a: float = -2.,
                  b: float = 2.) -> Tensor:
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def constant_(tensor: Tensor, val: float) -> Tensor:
    r"""Fills the input Tensor with the value :math:`\text{val}`.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        val: the value to fill the tensor with

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.constant_(w, 0.3)
    """
    return _no_grad_fill_(tensor, val)


def ones_(tensor: Tensor) -> Tensor:
    r"""Fills the input Tensor with the scalar value `1`.

    Args:
        tensor: an n-dimensional `torch.Tensor`

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.ones_(w)
    """
    return _no_grad_fill_(tensor, 1.)


def zeros_(tensor: Tensor) -> Tensor:
    r"""Fills the input Tensor with the scalar value `0`.

    Args:
        tensor: an n-dimensional `torch.Tensor`

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.zeros_(w)
    """
    return _no_grad_zero_(tensor)


def eye_(tensor):
    r"""Fills the 2-dimensional input `Tensor` with the identity
    matrix. Preserves the identity of the inputs in `Linear` layers, where as
    many inputs are preserved as possible.

    Args:
        tensor: a 2-dimensional `torch.Tensor`

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.eye_(w)
    """
    if tensor.ndimension() != 2:
        raise ValueError("Only tensors with 2 dimensions are supported")

    with paddle.no_grad():
        tensor.set_value(paddle.eye(*tensor.shape))
    return tensor


def dirac_(tensor, groups=1):
    r"""Fills the {3, 4, 5}-dimensional input `Tensor` with the Dirac
    delta function. Preserves the identity of the inputs in `Convolutional`
    layers, where as many input channels are preserved as possible. In case
    of groups>1, each group of channels preserves identity

    Args:
        tensor: a {3, 4, 5}-dimensional `torch.Tensor`
        groups (optional): number of groups in the conv layer (default: 1)
    Examples:
        >>> w = torch.empty(3, 16, 5, 5)
        >>> nn.init.dirac_(w)
        >>> w = torch.empty(3, 24, 5, 5)
        >>> nn.init.dirac_(w, 3)
    """
    dimensions = tensor.ndimension()
    if dimensions not in [3, 4, 5]:
        raise ValueError(
            "Only tensors with 3, 4, or 5 dimensions are supported")

    sizes = tensor.shape

    if sizes[0] % groups != 0:
        raise ValueError('dim 0 must be divisible by groups')

    out_chans_per_grp = sizes[0] // groups
    min_dim = min(out_chans_per_grp, sizes[1])

    with paddle.no_grad():
        tensor.zero_()

        for g in range(groups):
            for d in range(min_dim):
                if dimensions == 3:  # Temporal convolution
                    tensor[g * out_chans_per_grp + d, d,
                           tensor.shape[2] // 2] = 1
                elif dimensions == 4:  # Spatial convolution
                    tensor[g * out_chans_per_grp + d, d, tensor.shape[2] // 2,
                           tensor.shape[3] // 2] = 1
                else:  # Volumetric convolution
                    tensor[g * out_chans_per_grp + d, d, tensor.shape[2] // 2,
                           tensor.shape[3] // 2, tensor.shape[4] // 2] = 1
    return tensor


def _calculate_fan_in_and_fan_out(tensor):
    dimensions = tensor.dim()
    if dimensions < 2:
        raise ValueError(
            "Fan in and fan out can not be computed for tensor with fewer than 2 dimensions"
        )

    num_input_fmaps = tensor.shape[1]  # .size(1)
    num_output_fmaps = tensor.shape[0]  # .size(0)
    receptive_field_size = 1
    if tensor.dim() > 2:
        for s in tensor.shape[2:]:
            receptive_field_size *= s  # fixed
    fan_in = num_input_fmaps * receptive_field_size
    fan_out = num_output_fmaps * receptive_field_size

    return fan_in, fan_out


def LongTensor(x):
    return paddle.to_tensor(x, dtype='int64')


def IntTensor(x):
    return paddle.to_tensor(x, dtype='int32')


def xavier_uniform_(tensor: Tensor, gain: float = 1.) -> Tensor:
    r"""Fills the input `Tensor` with values according to the method
    described in `Understanding the difficulty of training deep feedforward
    neural networks` - Glorot, X. & Bengio, Y. (2010), using a uniform
    distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-a, a)` where

    .. math::
        a = \text{gain} \times \sqrt{\frac{6}{\text{fan\_in} + \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        gain: an optional scaling factor

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.xavier_uniform_(w, gain=nn.init.calculate_gain('relu'))
    """
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
    a = math.sqrt(
        3.0) * std  # Calculate uniform bounds from standard deviation

    return _no_grad_uniform_(tensor, -a, a)


def xavier_normal_(tensor: Tensor, gain: float = 1.) -> Tensor:
    r"""Fills the input `Tensor` with values according to the method
    described in `Understanding the difficulty of training deep feedforward
    neural networks` - Glorot, X. & Bengio, Y. (2010), using a normal
    distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{N}(0, \text{std}^2)` where

    .. math::
        \text{std} = \text{gain} \times \sqrt{\frac{2}{\text{fan\_in} + \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        gain: an optional scaling factor

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.xavier_normal_(w)
    """
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))

    return _no_grad_normal_(tensor, 0., std)


def _calculate_correct_fan(tensor, mode):
    mode = mode.lower()
    valid_modes = ['fan_in', 'fan_out']
    if mode not in valid_modes:
        raise ValueError("Mode {} not supported, please use one of {}".format(
            mode, valid_modes))

    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    return fan_in if mode == 'fan_in' else fan_out


def kaiming_uniform_(tensor, a=0, mode='fan_in', nonlinearity='leaky_relu'):
    r"""Fills the input `Tensor` with values according to the method
    described in `Delving deep into rectifiers: Surpassing human-level
    performance on ImageNet classification` - He, K. et al. (2015), using a
    uniform distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-\text{bound}, \text{bound})` where

    .. math::
        \text{bound} = \text{gain} \times \sqrt{\frac{3}{\text{fan\_mode}}}

    Also known as He initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        a: the negative slope of the rectifier used after this layer (only
            used with ``'leaky_relu'``)
        mode: either ``'fan_in'`` (default) or ``'fan_out'``. Choosing ``'fan_in'``
            preserves the magnitude of the variance of the weights in the
            forward pass. Choosing ``'fan_out'`` preserves the magnitudes in the
            backwards pass.
        nonlinearity: the non-linear function (`nn.functional` name),
            recommended to use only with ``'relu'`` or ``'leaky_relu'`` (default).

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.kaiming_uniform_(w, mode='fan_in', nonlinearity='relu')
    """
    fan = _calculate_correct_fan(tensor, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    bound = math.sqrt(
        3.0) * std  # Calculate uniform bounds from standard deviation
    with paddle.no_grad():
        tensor.set_value(paddle.uniform(tensor.shape, min=-bound, max=bound))
        return tensor


def kaiming_normal_(tensor, a=0, mode='fan_in', nonlinearity='leaky_relu'):
    r"""Fills the input `Tensor` with values according to the method
    described in `Delving deep into rectifiers: Surpassing human-level
    performance on ImageNet classification` - He, K. et al. (2015), using a
    normal distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{N}(0, \text{std}^2)` where

    .. math::
        \text{std} = \frac{\text{gain}}{\sqrt{\text{fan\_mode}}}

    Also known as He initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        a: the negative slope of the rectifier used after this layer (only
            used with ``'leaky_relu'``)
        mode: either ``'fan_in'`` (default) or ``'fan_out'``. Choosing ``'fan_in'``
            preserves the magnitude of the variance of the weights in the
            forward pass. Choosing ``'fan_out'`` preserves the magnitudes in the
            backwards pass.
        nonlinearity: the non-linear function (`nn.functional` name),
            recommended to use only with ``'relu'`` or ``'leaky_relu'`` (default).

    Examples:
        >>> w = torch.empty(3, 5)
        >>> kaiming_normal_(w, mode='fan_out', nonlinearity='relu')
    """
    fan = _calculate_correct_fan(tensor, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    with paddle.no_grad():
        tensor.set_value(paddle.normal(shape=tensor.shape, mean=0, std=std))
        return tensor


def orthogonal_(tensor, gain=1):
    r"""Fills the input `Tensor` with a (semi) orthogonal matrix, as
    described in `Exact solutions to the nonlinear dynamics of learning in deep
    linear neural networks` - Saxe, A. et al. (2013). The input tensor must have
    at least 2 dimensions, and for tensors with more than 2 dimensions the
    trailing dimensions are flattened.

    Args:
        tensor: an n-dimensional `torch.Tensor`, where :math:`n \geq 2`
        gain: optional scaling factor

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.orthogonal_(w)
    """
    if tensor.ndimension() < 2:
        raise ValueError(
            "Only tensors with 2 or more dimensions are supported")

    rows = tensor.shape[0]  # .size(0)
    cols = tensor.numel() // rows
    flattened = tensor.new(rows, cols).normal_(0, 1)

    if rows < cols:
        flattened.t_()

    # Compute the qr factorization
    q, r = paddle.to_tensor(np.linalg.qr(flattened.numpy()))
    # q, r = torch.qr(flattened)
    # Make Q uniform according to https://arxiv.org/pdf/math-ph/0609050.pdf
    d = paddle.diag(r, 0)
    ph = d.sign()
    q *= ph

    if rows < cols:
        q.t_()

    with paddle.no_grad():
        tensor.view_as(q).copy_(q)
        tensor.mul_(gain)
    return tensor


def sparse_(tensor, sparsity, std=0.01):
    r"""Fills the 2D input `Tensor` as a sparse matrix, where the
    non-zero elements will be drawn from the normal distribution
    :math:`\mathcal{N}(0, 0.01)`, as described in `Deep learning via
    Hessian-free optimization` - Martens, J. (2010).

    Args:
        tensor: an n-dimensional `torch.Tensor`
        sparsity: The fraction of elements in each column to be set to zero
        std: the standard deviation of the normal distribution used to generate
            the non-zero values

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.sparse_(w, sparsity=0.1)
    """
    if tensor.ndimension() != 2:
        raise ValueError("Only tensors with 2 dimensions are supported")

    rows, cols = tensor.shape
    num_zeros = int(math.ceil(sparsity * rows))

    with paddle.no_grad():
        tensor.normal_(0, std)
        for col_idx in range(cols):
            row_indices = paddle.randperm(rows)
            zero_indices = row_indices[:num_zeros]
            tensor[zero_indices, col_idx] = 0
    return tensor


# for backward compatibility
def _make_deprecate(meth):
    new_name = meth.__name__
    old_name = new_name[:-1]

    def deprecated_init(*args, **kwargs):
        warnings.warn(
            "nn.init.{} is now deprecated in favor of nn.init.{}.".format(
                old_name, new_name),
            stacklevel=2)
        return meth(*args, **kwargs)

    deprecated_init.__doc__ = r"""
    {old_name}(...)

    .. warning::
        This method is now deprecated in favor of :func:`torch.nn.init.{new_name}`.

    See :func:`~torch.nn.init.{new_name}` for details.""".format(
        old_name=old_name, new_name=new_name)
    deprecated_init.__name__ = old_name
    return deprecated_init


# uniform = _make_deprecate(uniform_)
# normal = _make_deprecate(normal_)
# constant = _make_deprecate(constant_)
# eye = _make_deprecate(eye_)
# dirac = _make_deprecate(dirac_)
# xavier_uniform = _make_deprecate(xavier_uniform_)
# xavier_normal = _make_deprecate(xavier_normal_)
# kaiming_uniform = _make_deprecate(kaiming_uniform_)
# kaiming_normal = _make_deprecate(kaiming_normal_)
# orthogonal = _make_deprecate(orthogonal_)
# sparse = _make_deprecate(sparse_)



import numpy as np
import paddle
import paddle.nn as nn

import paddle.nn.functional as F
from EIVideo.paddlevideo.utils.manet_utils import int_, float_, long_, load
from EIVideo.paddlevideo.utils.manet_utils import kaiming_normal_

#############################################################GLOBAL_DIST_MAP

MODEL_UNFOLD = True
WRONG_LABEL_PADDING_DISTANCE = 1e20


def _pairwise_distances(x, y, ys=None):
    """Computes pairwise squared l2 distances between tensors x and y.
    Args:
    x: Tensor of shape [n, feature_dim].
    y: Tensor of shape [m, feature_dim].
    Returns:
    Float32 distances tensor of shape [n, m].
    """

    xs = paddle.sum(x * x, 1)
    xs = xs.unsqueeze(1)
    if ys is None:
        ys = paddle.sum(y * y, 1)
        ys = ys.unsqueeze(0)
    else:
        ys = ys
    d = xs + ys - 2. * paddle.matmul(x, paddle.t(y))
    return d, ys


##################
def _flattened_pairwise_distances(reference_embeddings, query_embeddings, ys):
    """Calculates flattened tensor of pairwise distances between ref and query.
    Args:
    reference_embeddings: Tensor of shape [..., embedding_dim],
      the embedding vectors for the reference frame
    query_embeddings: Tensor of shape [n_query_images, height, width,
      embedding_dim], the embedding vectors for the query frames.
    Returns:
    A distance tensor of shape [reference_embeddings.size / embedding_dim,
    query_embeddings.size / embedding_dim]
    """
    embedding_dim = query_embeddings.shape[-1]
    reference_embeddings = reference_embeddings.reshape([-1, embedding_dim])
    first_dim = -1
    query_embeddings = query_embeddings.reshape([first_dim, embedding_dim])
    dists, ys = _pairwise_distances(query_embeddings, reference_embeddings, ys)
    return dists, ys


def _nn_features_per_object_for_chunk(reference_embeddings, query_embeddings,
                                      wrong_label_mask, k_nearest_neighbors,
                                      ys):
    """Extracts features for each object using nearest neighbor attention.
  Args:
    reference_embeddings: Tensor of shape [n_chunk, embedding_dim],
      the embedding vectors for the reference frame.
    query_embeddings: Tensor of shape [m_chunk, embedding_dim], the embedding
      vectors for the query frames.
    wrong_label_mask:
    k_nearest_neighbors: Integer, the number of nearest neighbors to use.
  Returns:
    nn_features: A float32 tensor of nearest neighbor features of shape
      [m_chunk, n_objects, feature_dim].
    """
    #    reference_embeddings_key = reference_embeddings
    #    query_embeddings_key = query_embeddings
    dists, ys = _flattened_pairwise_distances(reference_embeddings,
                                              query_embeddings, ys)

    dists = (paddle.unsqueeze(dists, 1) +
             paddle.unsqueeze(float_(wrong_label_mask), 0) *
             WRONG_LABEL_PADDING_DISTANCE)
    if k_nearest_neighbors == 1:
        features = paddle.min(dists, 2, keepdim=True)
    else:
        dists, _ = paddle.topk(-dists, k=k_nearest_neighbors, axis=2)
        dists = -dists
        valid_mask = (dists < WRONG_LABEL_PADDING_DISTANCE)
        masked_dists = dists * valid_mask.float()
        pad_dist = paddle.max(masked_dists, axis=2, keepdim=True)[0].tile(
            (1, 1, masked_dists.shape[-1]))
        dists = paddle.where(valid_mask, dists, pad_dist)
        # take mean of distances
        features = paddle.mean(dists, axis=2, keepdim=True)

    return features, ys


###
def _selected_pixel(ref_labels_flat, ref_emb_flat):
    index_list = paddle.arange(len(ref_labels_flat))
    index_list = index_list
    index_ = paddle.masked_select(index_list, ref_labels_flat != -1)

    index_ = long_(index_)
    ref_labels_flat = paddle.index_select(ref_labels_flat, index_, 0)
    ref_emb_flat = paddle.index_select(ref_emb_flat, index_, 0)

    return ref_labels_flat, ref_emb_flat


###


def _nearest_neighbor_features_per_object_in_chunks(reference_embeddings_flat,
                                                    query_embeddings_flat,
                                                    reference_labels_flat,
                                                    ref_obj_ids,
                                                    k_nearest_neighbors,
                                                    n_chunks, **cfg):
    """Calculates the nearest neighbor features per object in chunks to save mem.
    Uses chunking to bound the memory use.
    Args:
    reference_embeddings_flat: Tensor of shape [n, embedding_dim],
      the embedding vectors for the reference frame.
    query_embeddings_flat: Tensor of shape [m, embedding_dim], the embedding
      vectors for the query frames.
    reference_labels_flat: Tensor of shape [n], the class labels of the
      reference frame.
    ref_obj_ids: int tensor of unique object ids in the reference labels.
    k_nearest_neighbors: Integer, the number of nearest neighbors to use.
    n_chunks: Integer, the number of chunks to use to save memory
      (set to 1 for no chunking).
    Returns:
    nn_features: A float32 tensor of nearest neighbor features of shape
      [m, n_objects, feature_dim].
    """

    # reference_embeddings_flat = reference_embeddings_flat.cpu()
    # query_embeddings_flat = query_embeddings_flat.cpu()
    # reference_labels_flat = reference_labels_flat.cpu()
    # ref_obj_ids = ref_obj_ids.cpu()

    chunk_size = int_(
        np.ceil((float_(query_embeddings_flat.shape[0]) / n_chunks).numpy()))
    if cfg.get('test_mode'):
        reference_labels_flat, reference_embeddings_flat = _selected_pixel(
            reference_labels_flat, reference_embeddings_flat)
    wrong_label_mask = (reference_labels_flat != paddle.unsqueeze(
        ref_obj_ids, 1))
    all_features = []
    for n in range(n_chunks):
        if n == 0:
            ys = None
        if n_chunks == 1:
            query_embeddings_flat_chunk = query_embeddings_flat
        else:
            chunk_start = n * chunk_size
            chunk_end = (n + 1) * chunk_size
            query_embeddings_flat_chunk = query_embeddings_flat[
                chunk_start:chunk_end]
        features, ys = _nn_features_per_object_for_chunk(
            reference_embeddings_flat, query_embeddings_flat_chunk,
            wrong_label_mask, k_nearest_neighbors, ys)
        all_features.append(features)
    if n_chunks == 1:
        nn_features = all_features[0]
    else:
        nn_features = paddle.concat(all_features, axis=0)
    return nn_features


def nearest_neighbor_features_per_object(reference_embeddings,
                                         query_embeddings,
                                         reference_labels,
                                         k_nearest_neighbors,
                                         gt_ids=None,
                                         n_chunks=100,
                                         **cfg):
    """Calculates the distance to the nearest neighbor per object.
    For every pixel of query_embeddings calculate the distance to the
    nearest neighbor in the (possibly subsampled) reference_embeddings per object.
    Args:
    reference_embeddings: Tensor of shape [height, width, embedding_dim],
      the embedding vectors for the reference frame.
    query_embeddings: Tensor of shape [n_query_images, height, width,
      embedding_dim], the embedding vectors for the query frames.
    reference_labels: Tensor of shape [height, width, 1], the class labels of
      the reference frame.
    max_neighbors_per_object: Integer, the maximum number of candidates
      for the nearest neighbor query per object after subsampling,
      or 0 for no subsampling.
    k_nearest_neighbors: Integer, the number of nearest neighbors to use.
    gt_ids: Int tensor of shape [n_objs] of the sorted unique ground truth
      ids in the first frame. If None, it will be derived from
      reference_labels.
    n_chunks: Integer, the number of chunks to use to save memory
      (set to 1 for no chunking).
    Returns:
    nn_features: A float32 tensor of nearest neighbor features of shape
      [n_query_images, height, width, n_objects, feature_dim].
    gt_ids: An int32 tensor of the unique sorted object ids present
      in the reference labels.
    """
    # reference_embeddings = reference_embeddings.detach().cpu()
    # query_embeddings = query_embeddings.detach().cpu()
    # reference_labels = reference_labels.detach().cpu()

    assert (reference_embeddings.shape[:2] == reference_labels.shape[:2])
    h, w, _ = query_embeddings.shape
    reference_labels_flat = reference_labels.reshape([-1])
    if gt_ids is None:
        ref_obj_ids = paddle.unique(reference_labels_flat)[-1]
        ref_obj_ids = np.arange(0, ref_obj_ids + 1)
        gt_ids = paddle.to_tensor(ref_obj_ids)
        gt_ids = int_(gt_ids)
    else:
        gt_ids = int_(paddle.arange(0, gt_ids + 1))

    embedding_dim = query_embeddings.shape[-1]
    query_embeddings_flat = query_embeddings.reshape([-1, embedding_dim])
    reference_embeddings_flat = reference_embeddings.reshape(
        [-1, embedding_dim])
    nn_features = _nearest_neighbor_features_per_object_in_chunks(
        reference_embeddings_flat, query_embeddings_flat,
        reference_labels_flat, gt_ids, k_nearest_neighbors, n_chunks, **cfg)
    nn_features_dim = nn_features.shape[-1]
    nn_features = nn_features.reshape(
        [1, h, w, gt_ids.shape[0], nn_features_dim])
    return nn_features.cuda(), gt_ids


########################################################################LOCAL_DIST_MAP


def local_pairwise_distances2(x, y, max_distance=9):
    """Computes pairwise squared l2 distances using a local search window.
    Naive implementation using map_fn.
    Used as a slow fallback for when correlation_cost is not available.
    Args:
    x: Float32 tensor of shape [height, width, feature_dim].
    y: Float32 tensor of shape [height, width, feature_dim].
    max_distance: Integer, the maximum distance in pixel coordinates
      per dimension which is considered to be in the search window.
    Returns:
    Float32 distances tensor of shape
      [height, width, (2 * max_distance + 1) ** 2].
    """
    ori_h, ori_w, _ = x.shape
    x = paddle.transpose(x, [2, 0, 1]).unsqueeze(0)
    x = F.avg_pool2d(x, (2, 2), (2, 2))
    y = paddle.transpose(y, [2, 0, 1]).unsqueeze(0)
    y = F.avg_pool2d(y, (2, 2), (2, 2))

    _, channels, height, width = x.shape
    padding_val = 1e20
    padded_y = F.pad(y,
                     (max_distance, max_distance, max_distance, max_distance),
                     mode='constant',
                     value=padding_val)
    offset_y = F.unfold(padded_y, kernel_sizes=[height, width]).reshape(
        [1, channels, height, width, -1])
    x = x.reshape([1, channels, height, width, 1])
    minus = x - offset_y
    dists = paddle.sum(paddle.multiply(minus, minus),
                       axis=1).reshape([1, height, width,
                                        -1]).transpose([0, 3, 1, 2])
    dists = (paddle.nn.functional.sigmoid(dists) - 0.5) * 2
    dists = F.interpolate(dists,
                          size=[ori_h, ori_w],
                          mode='bilinear',
                          align_corners=True)
    dists = dists.squeeze(0).transpose([1, 2, 0])
    return dists


def local_previous_frame_nearest_neighbor_features_per_object(
        prev_frame_embedding,
        query_embedding,
        prev_frame_labels,
        gt_ids,
        max_distance=12):
    """Computes nearest neighbor features while only allowing local matches.
  Args:
    prev_frame_embedding: Tensor of shape [height, width, embedding_dim],
      the embedding vectors for the last frame.
    query_embedding: Tensor of shape [height, width, embedding_dim],
      the embedding vectors for the query frames.
    prev_frame_labels: Tensor of shape [height, width, 1], the class labels of
      the previous frame.
    gt_ids: Int Tensor of shape [n_objs] of the sorted unique ground truth
      ids in the first frame.
    max_distance: Integer, the maximum distance allowed for local matching.
  Returns:
    nn_features: A float32 np.array of nearest neighbor features of shape
      [1, height, width, n_objects, 1].
    """
    #     print(query_embedding.shape, prev_frame_embedding.shape)
    #     print(query_embedding.place, prev_frame_embedding.place)
    #     query_embedding = query_embedding.cpu()
    #     prev_frame_embedding = prev_frame_embedding.cpu()
    #     prev_frame_labels = prev_frame_labels.cpu()
    #     print(prev_frame_labels.place, prev_frame_embedding.place, query_embedding.place)

    d = local_pairwise_distances2(query_embedding,
                                  prev_frame_embedding,
                                  max_distance=max_distance)
    height, width = prev_frame_embedding.shape[:2]

    if MODEL_UNFOLD:

        labels = float_(prev_frame_labels).transpose([2, 0, 1]).unsqueeze(0)
        padded_labels = F.pad(labels, (
            2 * max_distance,
            2 * max_distance,
            2 * max_distance,
            2 * max_distance,
        ))
        offset_labels = F.unfold(padded_labels,
                                 kernel_sizes=[height, width],
                                 strides=[2,
                                          2]).reshape([height, width, -1, 1])
        offset_masks = paddle.equal(
            offset_labels,
            float_(gt_ids).unsqueeze(0).unsqueeze(0).unsqueeze(0))
    else:

        masks = paddle.equal(prev_frame_labels,
                             gt_ids.unsqueeze(0).unsqueeze(0))
        padded_masks = nn.functional.pad(masks, (
            0,
            0,
            max_distance,
            max_distance,
            max_distance,
            max_distance,
        ))
        offset_masks = []
        for y_start in range(2 * max_distance + 1):
            y_end = y_start + height
            masks_slice = padded_masks[y_start:y_end]
            for x_start in range(2 * max_distance + 1):
                x_end = x_start + width
                offset_mask = masks_slice[:, x_start:x_end]
                offset_masks.append(offset_mask)
        offset_masks = paddle.stack(offset_masks, axis=2)

    d_tiled = d.unsqueeze(-1).tile((1, 1, 1, gt_ids.shape[0]))
    pad = paddle.ones_like(d_tiled)
    d_masked = paddle.where(offset_masks, d_tiled, pad)
    dists = paddle.min(d_masked, axis=2)
    dists = dists.reshape([1, height, width, gt_ids.shape[0], 1])

    return dists


##############################################################


#################
class _res_block(nn.Layer):
    def __init__(self, in_dim, out_dim, **cfg):
        super(_res_block, self).__init__()
        self.conv1 = nn.Conv2D(in_dim,
                               out_dim,
                               kernel_size=3,
                               stride=1,
                               padding=1)
        self.relu1 = nn.ReLU()
        self.bn1 = paddle.nn.BatchNorm2D(out_dim, momentum=cfg['train_bn_mom'])
        self.conv2 = nn.Conv2D(out_dim,
                               out_dim,
                               kernel_size=3,
                               stride=1,
                               padding=1)
        self.relu2 = nn.ReLU()
        self.bn2 = paddle.nn.BatchNorm2D(out_dim, momentum=cfg['train_bn_mom'])

    def forward(self, x):
        res = x
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        x += res
        return x


####################
class IntSegHead(nn.Layer):
    def __init__(self, in_dim, emb_dim, **cfg):
        super(IntSegHead, self).__init__()
        self.conv1 = nn.Conv2D(in_dim,
                               emb_dim,
                               kernel_size=7,
                               stride=1,
                               padding=3)
        self.bn1 = paddle.nn.BatchNorm2D(emb_dim, momentum=cfg['train_bn_mom'])
        self.relu1 = nn.ReLU(True)
        self.res1 = _res_block(emb_dim, emb_dim, **cfg)
        self.res2 = _res_block(emb_dim, emb_dim, **cfg)
        self.conv2 = nn.Conv2D(256,
                               emb_dim,
                               kernel_size=3,
                               stride=1,
                               padding=1)
        self.bn2 = paddle.nn.BatchNorm2D(emb_dim, momentum=cfg['train_bn_mom'])
        self.relu2 = nn.ReLU(True)
        self.conv3 = nn.Conv2D(emb_dim, 1, 1, 1)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.res1(x)
        x = self.res2(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        x = self.conv3(x)
        return x


class _split_separable_conv2d(nn.Layer):
    def __init__(self, in_dim, out_dim, kernel_size=7, **cfg):
        super(_split_separable_conv2d, self).__init__()
        self.conv1 = nn.Conv2D(in_dim,
                               in_dim,
                               kernel_size=kernel_size,
                               stride=1,
                               padding=int((kernel_size - 1) / 2),
                               groups=in_dim)
        self.relu1 = nn.ReLU(True)
        self.bn1 = paddle.nn.BatchNorm2D(in_dim, momentum=cfg['train_bn_mom'])
        self.conv2 = nn.Conv2D(in_dim, out_dim, kernel_size=1, stride=1)
        self.relu2 = nn.ReLU(True)
        self.bn2 = paddle.nn.BatchNorm2D(out_dim, momentum=cfg['train_bn_mom'])
        kaiming_normal_(self.conv1.weight, mode='fan_out', nonlinearity='relu')
        kaiming_normal_(self.conv2.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu2(x)
        return x


class DynamicSegHead(nn.Layer):
    def __init__(self, in_dim, embed_dim, **cfg):
        super(DynamicSegHead, self).__init__()
        self.layer1 = _split_separable_conv2d(in_dim, embed_dim, **cfg)
        self.layer2 = _split_separable_conv2d(embed_dim, embed_dim, **cfg)
        self.layer3 = _split_separable_conv2d(embed_dim, embed_dim, **cfg)
        self.layer4 = _split_separable_conv2d(embed_dim, embed_dim, **cfg)
        self.conv = nn.Conv2D(embed_dim, 1, 1, 1)
        kaiming_normal_(self.conv.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.conv(x)
        return x


from ..registry import HEADS
"""
覆盖原理
class c1:
    def __init__(self):
        self.a = 1


class c2(c1):
    def __init__(self):
        super(c2, self).__init__()
        self.a = 2


c = c2()
print(c.a)

"""


@HEADS.register()
class IntVOS(nn.Layer):
    def __init__(self, feature_extracter, **cfg):
        super(IntVOS, self).__init__()
        self.feature_extracter = feature_extracter  ##embedding extractor
        self.feature_extracter.cls_conv = nn.Sequential()
        self.feature_extracter.upsample4 = nn.Sequential()
        self.semantic_embedding = None
        self.seperate_conv = nn.Conv2D(cfg['model_aspp_outdim'],
                                       cfg['model_aspp_outdim'],
                                       kernel_size=3,
                                       stride=1,
                                       padding=1,
                                       groups=cfg['model_aspp_outdim'])
        self.bn1 = paddle.nn.BatchNorm2D(cfg['model_aspp_outdim'],
                                         momentum=cfg['train_bn_mom'])
        self.relu1 = nn.ReLU(True)
        self.embedding_conv = nn.Conv2D(cfg['model_aspp_outdim'],
                                        cfg['model_semantic_embedding_dim'], 1,
                                        1)
        self.relu2 = nn.ReLU(True)
        self.bn2 = paddle.nn.BatchNorm2D(cfg['model_semantic_embedding_dim'],
                                         momentum=cfg['train_bn_mom'])
        self.semantic_embedding = nn.Sequential(*[
            self.seperate_conv, self.bn1, self.relu1, self.embedding_conv,
            self.bn2, self.relu2
        ])

        for m in self.semantic_embedding:
            if isinstance(m, nn.Conv2D):
                kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')

        self.dynamic_seghead = DynamicSegHead(
            in_dim=cfg['model_semantic_embedding_dim'] + 3,
            embed_dim=cfg['model_head_embedding_dim'],
            **cfg)  # propagation segm head
        if cfg['model_useintseg']:
            self.inter_seghead = IntSegHead(
                in_dim=cfg['model_semantic_embedding_dim'] + 3,
                emb_dim=cfg['model_head_embedding_dim'],
                **cfg)
        else:
            self.inter_seghead = DynamicSegHead(
                in_dim=cfg['model_semantic_embedding_dim'] + 2,
                embed_dim=cfg['model_head_embedding_dim'],
                **cfg)  # interaction segm head
        self.pretrained = cfg.get('pretrained', None)
        self.cfg = cfg

    def init_weights(self):
        if isinstance(self.pretrained, str) and self.pretrained.strip() != "":
            self.set_state_dict(load(self.pretrained, self.state_dict()))
            print('loaded pretrained model')

    def loss(self, **kwargs):
        return self.loss_func(**kwargs)

    def forward(self,
                x=None,
                ref_scribble_label=None,
                previous_frame_mask=None,
                normalize_nearest_neighbor_distances=True,
                use_local_map=True,
                seq_names=None,
                gt_ids=None,
                k_nearest_neighbors=1,
                global_map_tmp_dic=None,
                local_map_dics=None,
                interaction_num=None,
                start_annotated_frame=None,
                frame_num=None):

        x = self.extract_feature(x)
        #         print('extract_feature:', x.mean().item())
        ref_frame_embedding, previous_frame_embedding, current_frame_embedding = paddle.split(
            x, num_or_sections=3, axis=0)

        if global_map_tmp_dic is None:
            dic = self.prop_seghead(
                ref_frame_embedding,
                previous_frame_embedding,
                current_frame_embedding,
                ref_scribble_label,
                previous_frame_mask,
                normalize_nearest_neighbor_distances,
                use_local_map,
                seq_names,
                gt_ids,
                k_nearest_neighbors,
                global_map_tmp_dic,
                local_map_dics,
                interaction_num,
                start_annotated_frame,
                frame_num,
                self.dynamic_seghead,
            )
            return dic

        else:
            dic, global_map_tmp_dic = self.prop_seghead(
                ref_frame_embedding,
                previous_frame_embedding,
                current_frame_embedding,
                ref_scribble_label,
                previous_frame_mask,
                normalize_nearest_neighbor_distances,
                use_local_map,
                seq_names,
                gt_ids,
                k_nearest_neighbors,
                global_map_tmp_dic,
                local_map_dics,
                interaction_num,
                start_annotated_frame,
                frame_num,
                self.dynamic_seghead,
            )
            return dic, global_map_tmp_dic

    def extract_feature(self, x):
        x = self.feature_extracter(x)
        x = self.semantic_embedding(x)
        return x

    def prop_seghead(
        self,
        ref_frame_embedding=None,
        previous_frame_embedding=None,
        current_frame_embedding=None,
        ref_scribble_label=None,
        previous_frame_mask=None,
        normalize_nearest_neighbor_distances=True,
        use_local_map=True,
        seq_names=None,
        gt_ids=None,
        k_nearest_neighbors=1,
        global_map_tmp_dic=None,
        local_map_dics=None,
        interaction_num=None,
        start_annotated_frame=None,
        frame_num=None,
        dynamic_seghead=None,
    ):
        """return: feature_embedding,global_match_map,local_match_map,previous_frame_mask"""
        ###############
        cfg = self.cfg
        global_map_tmp_dic = global_map_tmp_dic
        dic_tmp = {}
        bs, c, h, w = current_frame_embedding.shape
        if cfg.get('test_mode'):
            scale_ref_scribble_label = float_(ref_scribble_label)
        else:
            scale_ref_scribble_label = paddle.nn.functional.interpolate(
                float_(ref_scribble_label), size=(h, w), mode='nearest')
        scale_ref_scribble_label = int_(scale_ref_scribble_label)
        scale_previous_frame_label = paddle.nn.functional.interpolate(
            float_(previous_frame_mask), size=(h, w), mode='nearest')
        scale_previous_frame_label = int_(scale_previous_frame_label)
        for n in range(bs):
            seq_current_frame_embedding = current_frame_embedding[n]
            seq_ref_frame_embedding = ref_frame_embedding[n]
            seq_prev_frame_embedding = previous_frame_embedding[n]
            seq_ref_frame_embedding = seq_ref_frame_embedding.transpose(
                [1, 2, 0])
            seq_current_frame_embedding = seq_current_frame_embedding.transpose(
                [1, 2, 0])
            seq_ref_scribble_label = scale_ref_scribble_label[n].transpose(
                [1, 2, 0])
            #########Global Map
            nn_features_n, ref_obj_ids = nearest_neighbor_features_per_object(
                reference_embeddings=seq_ref_frame_embedding,
                query_embeddings=seq_current_frame_embedding,
                reference_labels=seq_ref_scribble_label,
                k_nearest_neighbors=k_nearest_neighbors,
                gt_ids=gt_ids[n],
                n_chunks=10)
            if normalize_nearest_neighbor_distances:
                nn_features_n = (paddle.nn.functional.sigmoid(nn_features_n) -
                                 0.5) * 2

            #             print(nn_features_n)

            ###
            if global_map_tmp_dic is not None:  ###when testing, use global map memory
                if seq_names[n] not in global_map_tmp_dic:
                    global_map_tmp_dic[seq_names[n]] = paddle.ones_like(
                        nn_features_n).tile([1000, 1, 1, 1, 1])
                nn_features_n = paddle.where(
                    nn_features_n <= global_map_tmp_dic[seq_names[n]][
                        frame_num[n]].unsqueeze(0), nn_features_n,
                    global_map_tmp_dic[seq_names[n]][frame_num[n]].unsqueeze(
                        0))

                #                 print('detach 1')
                #                 print(nn_features_n.shape)
                # nn_features_n = nn_features_n.detach()
                global_map_tmp_dic[seq_names[n]][
                    frame_num[n]] = nn_features_n.detach()[0]

            #########################Local dist map
            seq_prev_frame_embedding = seq_prev_frame_embedding.transpose(
                [1, 2, 0])
            seq_previous_frame_label = scale_previous_frame_label[n].transpose(
                [1, 2, 0])

            if use_local_map:
                prev_frame_nn_features_n = local_previous_frame_nearest_neighbor_features_per_object(
                    prev_frame_embedding=seq_prev_frame_embedding,
                    query_embedding=seq_current_frame_embedding,
                    prev_frame_labels=seq_previous_frame_label,
                    gt_ids=ref_obj_ids,
                    max_distance=cfg['model_max_local_distance'])
            else:
                prev_frame_nn_features_n, _ = nearest_neighbor_features_per_object(
                    reference_embeddings=seq_prev_frame_embedding,
                    query_embeddings=seq_current_frame_embedding,
                    reference_labels=seq_previous_frame_label,
                    k_nearest_neighbors=k_nearest_neighbors,
                    gt_ids=gt_ids[n],
                    n_chunks=20)
                prev_frame_nn_features_n = (
                    paddle.nn.functional.sigmoid(prev_frame_nn_features_n) -
                    0.5) * 2

            #             print(prev_frame_nn_features_n.mean().item(), prev_frame_nn_features_n.shape, interaction_num)  # o
            #############
            if local_map_dics is not None:  ##When testing, use local map memory
                local_map_tmp_dic, local_map_dist_dic = local_map_dics
                if seq_names[n] not in local_map_dist_dic:
                    print(seq_names[n], 'not in local_map_dist_dic')
                    local_map_dist_dic[seq_names[n]] = paddle.zeros(1000, 9)
                if seq_names[n] not in local_map_tmp_dic:
                    print(seq_names[n], 'not in local_map_tmp_dic')
                    local_map_tmp_dic[seq_names[n]] = paddle.zeros_like(
                        prev_frame_nn_features_n).unsqueeze(0).tile(
                            [1000, 9, 1, 1, 1, 1])
                #                 print(local_map_dist_dic[seq_names[n]].shape)
                #                 print('detach 2')
                # prev_frame_nn_features_n = prev_frame_nn_features_n.detach()
                local_map_dist_dic[seq_names[n]][
                    frame_num[n], interaction_num -
                    1] = 1.0 / (abs(frame_num[n] - start_annotated_frame)
                                )  # bugs fixed.
                local_map_tmp_dic[seq_names[n]][
                    frame_num[n],
                    interaction_num - 1] = prev_frame_nn_features_n.squeeze(
                        0).detach()  # bugs fixed.
                if interaction_num == 1:
                    prev_frame_nn_features_n = local_map_tmp_dic[seq_names[n]][
                        frame_num[n]][interaction_num - 1]
                    prev_frame_nn_features_n = prev_frame_nn_features_n.unsqueeze(
                        0)
                else:
                    if local_map_dist_dic[seq_names[n]][frame_num[n]][interaction_num - 1] > \
                            local_map_dist_dic[seq_names[n]][frame_num[n]][interaction_num - 2]:
                        prev_frame_nn_features_n = local_map_tmp_dic[
                            seq_names[n]][frame_num[n]][interaction_num - 1]
                        prev_frame_nn_features_n = prev_frame_nn_features_n.unsqueeze(
                            0)
                    else:
                        prev_frame_nn_features_n = local_map_tmp_dic[
                            seq_names[n]][frame_num[n]][interaction_num - 2]
                        prev_frame_nn_features_n = prev_frame_nn_features_n.unsqueeze(
                            0)

                local_map_dics = (local_map_tmp_dic, local_map_dist_dic)

            to_cat_previous_frame = (
                float_(seq_previous_frame_label) == float_(ref_obj_ids)
            )  # float comparision?

            to_cat_current_frame_embedding = current_frame_embedding[
                n].unsqueeze(0).tile((ref_obj_ids.shape[0], 1, 1, 1))

            to_cat_nn_feature_n = nn_features_n.squeeze(0).transpose(
                [2, 3, 0, 1])
            to_cat_previous_frame = float_(
                to_cat_previous_frame.unsqueeze(-1).transpose([2, 3, 0, 1]))
            to_cat_prev_frame_nn_feature_n = prev_frame_nn_features_n.squeeze(
                0).transpose([2, 3, 0, 1])
            to_cat = paddle.concat(
                (to_cat_current_frame_embedding, to_cat_nn_feature_n,
                 to_cat_prev_frame_nn_feature_n, to_cat_previous_frame), 1)
            pred_ = dynamic_seghead(to_cat)
            pred_ = pred_.transpose([1, 0, 2, 3])
            dic_tmp[seq_names[n]] = pred_

        if global_map_tmp_dic is None:
            return dic_tmp
        else:
            if local_map_dics is None:
                return dic_tmp, global_map_tmp_dic
            else:
                return dic_tmp, global_map_tmp_dic, local_map_dics

    def int_seghead(self,
                    ref_frame_embedding=None,
                    ref_scribble_label=None,
                    prev_round_label=None,
                    normalize_nearest_neighbor_distances=True,
                    global_map_tmp_dic=None,
                    local_map_dics=None,
                    interaction_num=None,
                    seq_names=None,
                    gt_ids=None,
                    k_nearest_neighbors=1,
                    frame_num=None,
                    first_inter=True):
        dic_tmp = {}
        bs, c, h, w = ref_frame_embedding.shape
        scale_ref_scribble_label = paddle.nn.functional.interpolate(
            float_(ref_scribble_label), size=(h, w), mode='nearest')
        scale_ref_scribble_label = int_(scale_ref_scribble_label)
        if not first_inter:
            scale_prev_round_label = paddle.nn.functional.interpolate(
                float_(prev_round_label), size=(h, w), mode='nearest')
            scale_prev_round_label = int_(scale_prev_round_label)
        n_chunks = 500
        for n in range(bs):

            gt_id = paddle.arange(0, gt_ids[n] + 1)

            gt_id = int_(gt_id)

            seq_ref_frame_embedding = ref_frame_embedding[n]

            ########################Local dist map
            seq_ref_frame_embedding = paddle.transpose(seq_ref_frame_embedding,
                                                       [1, 2, 0])
            seq_ref_scribble_label = paddle.transpose(
                scale_ref_scribble_label[n], [1, 2, 0])
            nn_features_n = local_previous_frame_nearest_neighbor_features_per_object(
                prev_frame_embedding=seq_ref_frame_embedding,
                query_embedding=seq_ref_frame_embedding,
                prev_frame_labels=seq_ref_scribble_label,
                gt_ids=gt_id,
                max_distance=self.cfg['model_max_local_distance'])

            #######
            ######################Global map update
            if seq_names[n] not in global_map_tmp_dic:
                global_map_tmp_dic[seq_names[n]] = paddle.ones_like(
                    nn_features_n).tile([1000, 1, 1, 1, 1])
            nn_features_n_ = paddle.where(
                nn_features_n <=
                global_map_tmp_dic[seq_names[n]][frame_num[n]].unsqueeze(0),
                nn_features_n,
                global_map_tmp_dic[seq_names[n]][frame_num[n]].unsqueeze(0))

            ###

            ###
            #             print('detach 3')
            # nn_features_n_ = nn_features_n_.detach()
            global_map_tmp_dic[seq_names[n]][
                frame_num[n]] = nn_features_n_.detach()[0]
            ##################Local map update
            if local_map_dics is not None:
                local_map_tmp_dic, local_map_dist_dic = local_map_dics
                if seq_names[n] not in local_map_dist_dic:
                    local_map_dist_dic[seq_names[n]] = paddle.zeros([1000, 9])
                if seq_names[n] not in local_map_tmp_dic:
                    local_map_tmp_dic[seq_names[n]] = paddle.ones_like(
                        nn_features_n).unsqueeze(0).tile([1000, 9, 1, 1, 1, 1])
                local_map_dist_dic[seq_names[n]][frame_num[n]][interaction_num
                                                               - 1] = 0

                local_map_dics = (local_map_tmp_dic, local_map_dist_dic)

            ##################
            to_cat_current_frame_embedding = ref_frame_embedding[n].unsqueeze(
                0).tile((gt_id.shape[0], 1, 1, 1))
            to_cat_nn_feature_n = nn_features_n.squeeze(0).transpose(
                [2, 3, 0, 1])

            to_cat_scribble_mask_to_cat = (
                float_(seq_ref_scribble_label) == float_(gt_id)
            )  # float comparision?
            to_cat_scribble_mask_to_cat = float_(
                to_cat_scribble_mask_to_cat.unsqueeze(-1).transpose(
                    [2, 3, 0, 1]))
            if not first_inter:
                seq_prev_round_label = scale_prev_round_label[n].transpose(
                    [1, 2, 0])

                to_cat_prev_round_to_cat = (
                    float_(seq_prev_round_label) == float_(gt_id)
                )  # float comparision?
                to_cat_prev_round_to_cat = float_(
                    to_cat_prev_round_to_cat.unsqueeze(-1).transpose(
                        [2, 3, 0, 1]))
            else:
                to_cat_prev_round_to_cat = paddle.zeros_like(
                    to_cat_scribble_mask_to_cat)
                to_cat_prev_round_to_cat[0] = 1.

            to_cat = paddle.concat(
                (to_cat_current_frame_embedding, to_cat_scribble_mask_to_cat,
                 to_cat_prev_round_to_cat), 1)

            pred_ = self.inter_seghead(to_cat)
            pred_ = pred_.transpose([1, 0, 2, 3])
            dic_tmp[seq_names[n]] = pred_
        if local_map_dics is None:
            return dic_tmp
        else:
            return dic_tmp, local_map_dics



# Author: Acer Zhang
# Datetime: 2022/1/11
# Copyright belongs to the author.
# Please indicate the source for reprinting.
from setuptools import setup, find_packages

with open('requirements.txt', encoding="utf-8-sig") as f:
    requirements = f.readlines()

setup(
    name='EIVideo',
    version="0.1a2.dev1",
    packages=find_packages(),
    long_description="[https://github.com/QPT-Family/EIVideo](https://github.com/QPT-Family/EIVideo)",
    long_description_content_type='text/markdown',
    url='https://github.com/QPT-Family/EIVideo',
    license='LGPL',
    author='GT-ZhangAcer',
    author_email='zhangacer@foxmail.com',
    description='EIVideo - 交互式智能视频标注工具，几次鼠标点击即可解放双手，让视频标注更加轻松',
    install_requires=requirements,
    python_requires='>=3.7',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'eivideo = eivideo.main:cli',
        ]}
)
